from typing import Optional

import numpy as np

from packg.typext import TensorType


def ensure_numpy(inp: TensorType) -> np.ndarray:
    if isinstance(inp, np.ndarray):
        return inp
    if hasattr(inp, "numpy"):
        return inp.numpy()
    return np.array(inp)


class AvgMetric:
    def __init__(self):
        self.reset()

    def reset(self):
        self.count = 0
        self.avg = 0.0

    def update(self, value: float, n: int = 1):
        if n <= 0:
            raise ValueError(f"n must be > 0 but is {n}")
        self.count += n
        self.avg = (self.avg * (self.count - n) + value * n) / self.count


def describe_stats(
    arr: TensorType,
    name: str = "",
    format_str: str = "{:.2f}",
    table_sep: Optional[str] = None,
) -> str:
    """
    See also lovely-tensors package for things like this
    see also pandas.describe

    Args:
        arr: array or tensor
        name: name of the tensor
        format_str: format string for printing
        table_sep: separator for table output (e.g. CSV) or none for human readable output (default)

    """
    arr = ensure_numpy(arr)
    out_strs = []
    if table_sep is not None:
        fmt_str = table_sep.join([format_str] * 5)
        out_strs.append(f"{name}{table_sep}{len(arr)}{table_sep}")
    else:
        fmt_str = (
            "Range: "
            + format_str
            + " to "
            + format_str
            + ", mean: "
            + format_str
            + ", std: "
            + format_str
            + ", median: "
            + format_str
        )
        out_strs.append(f"{name}: #{len(arr)}, ")
    out_strs.append(
        fmt_str.format(np.min(arr), np.max(arr), np.mean(arr), np.std(arr), np.median(arr))
    )
    return "".join(out_strs)



"""
Wrapper for tqdm progressbar with default settings.

Note about smoothing:
    Smoothing: 0=compute average speed, 1=compute current speed
    (math: EMAcurrent = (1 − smoothing) * EMAprevious + smoothing * valuecurrent)
    Default is 0.3
"""
from typing import Optional

from tqdm import tqdm

TQDM_WID = 90


class tqdm_max_ncols(tqdm):
    """
    Wrapper for tqdm progressbar with a maximum width. Default max_ncols=90.

    This allows resizing a window later on without breaking the progressbar output.
    Usually tqdm would resize dynamically, but this doesn't always work (e.g. when using screen).
    """

    def __init__(self, *args, max_ncols: Optional[int] = TQDM_WID, **kwargs):
        super().__init__(*args, **kwargs)
        if self.disable:
            # pbar is disabled, no need to modify ncols
            return
        if max_ncols is not None and self.ncols is not None:
            self.ncols = min(self.ncols, max_ncols)



"""
The logger provided here is a link to the loguru.logger but sets default level to ERROR.
To activate logging, use the configure_logger function:

    >>> configure_logger(level=logging.DEBUG)

Other contents of this file: Utils for the standard library logging package,
Utils for the loguru package.
https://loguru.readthedocs.io/en/stable/resources/recipes.html#changing-the-level-of-an-existing-handler
"""
from __future__ import annotations

from logging import getLevelName

import itertools
import logging
import os
import sys
from copy import deepcopy
from deprecated import deprecated
from loguru import logger as loguru_logger
from pathspec import PathSpec
from typing import Union

from packg.iotools.pathspec_matcher import make_pathspec
from typedparser import VerboseQuietArgs

LevelType = Union[str, int]  # either "DEBUG" or 10
DEFAULT_LOGURU_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - "
    "<level>{message}</level>"
)
SHORTER_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> "
    "<level>{message}</level>"
)
SHORTEST_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<level>{message}</level>"
)
BRIGHTBG_FORMAT = (
    "<green>{time:YYYYMMDD HH:mm:ss}</green> <level>{level: <4.4}</level> "
    "<blue>{name}</blue>:<blue>{function}</blue>:<blue>{line}</blue> "
    "<level>{message}</level>"
)
TIMELESS_FORMAT = "<level>{level: <4.4}</level> <level>{message}</level>"

SPINNER_STR = "|/-\\"
SPINNER_CYCLE = itertools.cycle("|/-\\")


def get_stdlib_logging_formatter():
    return logging.Formatter(
        "%(asctime)s %(levelname)-4s %(filename)s:%(lineno)d %(message)s",
        datefmt="%Y%m%d %H:%M:%S",
    )


global_logger_config = None


def configure_logger(
    level: LevelType = "DEBUG",
    sink=sys.stderr,
    format=SHORTEST_FORMAT,  # noqa # pylint: disable=redefined-builtin
    colorize=True,
    add_sinks: Union[list[any], list[dict[str, any]], None] = None,
    kwargs_handler: Union[dict[str, any], None] = None,
    **kwargs: any,
) -> dict[str, any]:
    """
    Configure the loguru logger. For more complex usages, use logger.configure() directly.

    Args:
        level: minimum level to log
        sink: where to write the logs to
        format: message formatting
        colorize: add color codes to the output
        add_sinks: list of other sinks to add. str sinks will write to file.
            use a dictionary to change parameters format, colorize, level separately from main sink.
        kwargs_handler: additional parameters to pass to each handler
        **kwargs: additional parameters to pass to logger.configure()

    References:
        https://loguru.readthedocs.io/en/stable/api/logger.html
        from loguru import _colorizer  # color code reference

    Returns:
        Configuration passed to logger.configure()

    """
    add_sinks = add_sinks if add_sinks is not None else []
    kwargs_handler = kwargs_handler if kwargs_handler is not None else {}
    sinks = [sink] + add_sinks
    handlers = []
    level_str = get_level_as_str(level)
    for lsink in sinks:
        if isinstance(lsink, dict):
            params = lsink
            if "format" not in params:
                params["format"] = format
            if "colorize" not in params:
                params["colorize"] = colorize
            if "level" not in params:
                params["level"] = level_str
        else:
            params = {
                "sink": lsink,
                "format": format,
                "colorize": colorize,
                "level": level_str,
            }
        assert "sink" not in kwargs_handler, f"sink is a reserved key, got {params}"
        params.update(deepcopy(kwargs_handler))
        if isinstance(lsink, str):
            # automatically disable color codes for files
            params["colorize"] = False
        handlers.append(params)

    # fix bug when combining colorama and loguru in conemu console on windows
    if (
        colorize
        and os.name == "nt"
        and os.environ.get("CONEMUANSI") == "ON"
        and sink in (sys.stderr, sys.stdout)
    ):
        print("\r", end="")  # the invisible print here magically fixes the color

    logger_config = {"handlers": handlers, **kwargs}
    loguru_logger.configure(**logger_config)
    global global_logger_config
    global_logger_config = logger_config
    return logger_config


def reroute_logger(new_sink=sys.stderr, logger_config=None, handler_num: int = 0) -> None:
    """
    Reroute a sink from one target to another. Useful for proper logging inside
    a tqdm progressbar without having to recreate the entire logger.

    Args:
        new_sink: new target
        logger_config: config created by configure_logger() function above.
            If None, use the global config.
        handler_num: index of the handler of which to change the sink

    Usage:
        logger_config = configure_logger(...)
        pbar = tqdm(...)
        reroute_logger(lambda msg: pbar.write(msg, end=""), logger_config)
        # use tqdm progressbar and logger together here
        pbar.close()
        reroute_logger(sys.stderr, logger_config)

    Notes:
        The config is modified inplace so does not need to be returned here.
        It cannot be deepcopy-ed here because sink is not pickle-able.
    """
    if logger_config is None:
        logger_config = global_logger_config
    logger_config["handlers"][handler_num]["sink"] = new_sink
    loguru_logger.configure(**logger_config)
    return logger_config


def get_level_as_str(level: LevelType):
    if isinstance(level, str):
        return level.upper()
    if isinstance(level, int):
        return getLevelName(level).upper()
    raise TypeError(f"Level must be str or int, not {type(level)}")


def get_level_as_int(level: LevelType):
    if isinstance(level, int):
        return level
    level_int = int(getLevelName(level.upper()))
    return level_int


def get_logger_level_from_args(args: VerboseQuietArgs) -> str:
    if args.verbose:
        return "DEBUG"
    if args.quiet:
        return "WARNING"
    return "INFO"


def get_terminal_size() -> int:
    try:
        n_cols = os.get_terminal_size().columns
    except OSError:
        n_cols = 80
    return n_cols


def configure_stdlib_base_logger(
    level: LevelType = "INFO",
    format="%(levelname)s: %(message)s",  # noqa  # pylint: disable=redefined-builtin
) -> None:
    """
    Configure the standard library logger.

    Args:
        level: minimum level to log
        format: message formatting
            (see https://docs.python.org/3/library/logging.html#logrecord-attributes)
    """
    level = get_level_as_int(level)
    logging.basicConfig(level=level, format=format)


def silence_stdlib_loggers(
    search_str: str,
    regex_mode: bool = False,
    level: LevelType = logging.ERROR,
    verbose: bool = False,
):
    spec: PathSpec = make_pathspec([search_str], regex_mode=regex_mode)
    # print([p.regex for p in spec.patterns])
    level = get_level_as_int(level)
    for name in logging.root.manager.loggerDict.keys():
        loggr = logging.getLogger(name)
        if spec.match_file(name):
            loggr.setLevel(level)
            if verbose:
                print(f"Set verbosity to {level} for logger '{name}'")


# the default loguru logger level is debug, but it should be error.
# use configure_logger function anywhere to change the logger level.
configure_logger(level=logging.ERROR, sink=sys.stderr, format=SHORTEST_FORMAT, colorize=True)
logger = loguru_logger
logger.warn = deprecated("Use logger.warning instead")(logger.warning)



"""
Download a list of URLs

Currently this uses the last part of the URL as a filename.
"""

import os
import time
from collections import Counter
from pathlib import Path
from typing import Optional

from loguru import logger

from packg.iotools import yield_lines_from_file
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.multiproc import FnMultiProcessor
from packg.web import download_file
from typedparser import VerboseQuietArgs, add_argument, TypedParser
from attrs import define


@define
class Args(VerboseQuietArgs):
    target_dir: Optional[Path] = add_argument(
        shortcut="-t", type=str, help="Target dir", required=True
    )
    input_list: Optional[Path] = add_argument(
        shortcut="-i", type=str, help="List of URLs", required=True
    )
    filename_list: Optional[Path] = add_argument(
        shortcut="-f", type=str, help="List of filenames", default=None
    )
    prefix: Optional[str] = add_argument(shortcut="-p", type=str, help="Prefix", default=None)
    cpus: int = add_argument(shortcut="-c", type=int, help="Number of cpus", default=0)
    sleep_time: float = add_argument(shortcut="-s", type=float, help="Sleep time", default=1)
    min_size_mb: float = add_argument(shortcut="-m", type=float, help="Min size in MB", default=0.0)
    n_retries: int = add_argument(shortcut="-r", type=int, help="Number of retries", default=3)


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    input_list = list(yield_lines_from_file(args.input_list))
    logger.info(f"Found {len(input_list)} lines in {args.input_list}")
    if args.prefix is not None:
        input_list = [f"{args.prefix}{line}" for line in input_list]
    logger.info(f"First 10 URLs: {input_list[:10]}")

    # load list of filenames if exists
    if args.filename_list is not None:
        filename_list = list(yield_lines_from_file(args.filename_list))
        assert len(filename_list) == len(
            input_list
        ), f"Filename list length {len(filename_list)} != URL list length {len(input_list)}"
        url_to_file = {url: file for url, file in zip(input_list, filename_list)}
    else:
        url_to_file = {url: Path(url).name for url in input_list}

    url_status = {url: "todo" for url in input_list}
    url_to_file_here = {}
    for url, file in url_to_file.items():
        full_file = args.target_dir / file
        if full_file.is_file():
            url_status[url] = "done"
            continue
        url_to_file_here[url] = full_file
    logger.info(f"Downloading status: {Counter(url_status.values())}")

    os.makedirs(args.target_dir, exist_ok=True)
    proc = FnMultiProcessor(
        args.cpus, download_fn, total=len(url_to_file_here), desc="Downloading URLs"
    )
    for url, file in url_to_file_here.items():
        if file.is_file():
            continue
        proc.put(file, url, args.sleep_time, args.min_size_mb, args.n_retries)
    proc.run()
    outputs = []
    for _ in range(len(url_to_file_here)):
        outputs.append(proc.get())
    proc.close()


def download_fn(file, url, sleep_time, min_size_mb, n_retries):
    n_tries = 0
    while True:
        success = True
        n_tries += 1
        try:
            download_file(file, url, verbose=False, pbar=False)
        except KeyboardInterrupt:
            # delete file to avoid having half files leftover
            if file.is_file():
                file.unlink()
        if min_size_mb > 0 and Path(file).stat().st_size < min_size_mb * 1024**2:
            logger.warning(
                f"File {file} is too small, retrying after sleep. "
                f"Got: {Path(file).read_text(encoding='utf-8')}"
            )
            os.unlink(file)
            success = False
        if success:
            break
        elif n_tries >= n_retries:
            logger.error(f"Too many retries for {url} {file}")
            break
        else:
            time.sleep(sleep_time)
            continue

    time.sleep(sleep_time)
    return True


if __name__ == "__main__":
    main()



"""
Wrapper functions for json files.

Behaviour of the JSON writer is changed to:
    - Automatically converts JSON incompatible data types to lists
    (pathlib.Path, numpy arrays, torch tensors, jax tensors) instead of raising errors
    - Does not indent each list element as a new line

Possible improvements:
    - Refactor the "file_or_io" handling into a context manager

"""
import io
import json
import os
from functools import partial
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Iterable, Sequence, List

from packg import format_exception
from packg.iotools.compress import (
    CompressorC,
    decompress_file_to_str,
    compress_data_to_file,
)
from packg.iotools.file_reader import (
    open_file_or_io,
    read_text_from_file_or_io,
)
from packg.iotools.jsonext_encoder import CustomJSONEncoder
from packg.typext import PathOrIO, PathTypeCls, PathType


def load_json(file_or_io: PathOrIO, verbose: bool = False, encoding: str = "utf-8") -> Any:
    """Load data from json file or file object"""
    start_timer = timer()
    if verbose:
        try:
            file_len = f"{Path(file_or_io).stat().st_size / 1024 ** 2:.1f} MB"
        except Exception:
            try:
                file_len = f"{len(file_or_io) / 1024 ** 2:.1f} M chars"
            except Exception:
                file_len = f"unknown"
        print(f"Load json file {file_or_io} with size {file_len}.")
    data_str = read_text_from_file_or_io(file_or_io, encoding=encoding)
    try:
        obj = loads_json(data_str)
    except Exception as e:
        # # todo use a general way to reraise the same exception with added information.
        # print(f"{e=} {e.args=}")
        # if len(e.args) > 0:
        #     _msg = e.args[0]
        #     remaining_args = e.args[1:]
        # else:
        #     remaining_args = tuple()
        # print(f"{remaining_args=} {type(e)=}")
        # raise type(e)(
        #     f"Error loading json file {file_or_io} due to {format_exception(e)}", *remaining_args
        # ) from e
        raise RuntimeError(
            f"Error loading json file {file_or_io} due to {format_exception(e)}"
        ) from e

    if verbose:
        print(f"Loaded json file {file_or_io} in {timer() - start_timer:.3f} seconds")
    return obj


def load_json_compressed(
    file_or_io: PathOrIO,
    compressor_name: CompressorC,
    verbose: bool = False,
    encoding: str = "utf-8",
    **compressor_kwargs,
) -> Any:
    start_timer = timer()
    data_text = decompress_file_to_str(file_or_io, compressor_name, encoding, **compressor_kwargs)
    try:
        obj = loads_json(data_text)
    except Exception as e:
        raise RuntimeError(f"Error loading compressed json file {file_or_io}") from e
    if verbose:
        print(f"Loaded json file {file_or_io} in {timer() - start_timer:.3f} seconds")
    return obj


def loads_json(s: str) -> Any:
    """Load data from json string"""
    return json.loads(s)


def load_jsonl(file_or_io: PathOrIO, encoding: str = "utf-8") -> List[Any]:
    """Load data from jsonl (list of json strings) file or file object.

    Notes:
        This is inefficient for large files. In that case, iterate the file lines
        and use loads_json on each line.
    """
    data_str = read_text_from_file_or_io(file_or_io, encoding=encoding)
    try:
        data = loads_jsonl(data_str)
    except Exception as e:
        raise RuntimeError(f"Error loading jsonl file {file_or_io}") from e
    return data


def load_jsonl_compressed(
    file_or_io: PathOrIO,
    compressor_name: CompressorC,
    verbose: bool = False,
    encoding: str = "utf-8",
    **compressor_kwargs,
) -> Any:
    start_timer = timer()
    data_text = decompress_file_to_str(file_or_io, compressor_name, encoding, **compressor_kwargs)
    try:
        obj = loads_jsonl(data_text)
    except Exception as e:
        raise RuntimeError(f"Error loading compressed jsonl file {file_or_io}") from e
    if verbose:
        print(f"Loaded json file {file_or_io} in {timer() - start_timer:.3f} seconds")
    return obj


def loads_jsonl(s: str) -> List[Any]:
    data = []
    for i, line in enumerate(s.splitlines()):
        try:
            data.append(loads_json(line))
        except Exception as e:
            raise RuntimeError(f"Error loading json line {i}: {line}") from e
    return data


def _check_can_write(file_or_io, overwrite, verbose):
    if not overwrite and isinstance(file_or_io, PathTypeCls) and Path(file_or_io).is_file():
        if verbose:
            print(f"File already exists and overwrite is False: {Path(file_or_io).as_posix()}")
        return False
    return True


def dump_json_safely(obj: Any, f: PathType, **kwargs):
    try:
        dump_json(obj, f, **kwargs)
    except KeyboardInterrupt as e:
        print(f"Deleting {f} due to KeyboardInterrupt")
        try:
            os.remove(f)
        except Exception as e2:
            print(f"Exception while deleting {f}: {format_exception(e2)}")
            pass
        raise e


def dump_json(
    obj: Any,
    file_or_io: PathOrIO,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    verbose: bool = True,
    create_parent=False,
    float_precision=None,
    custom_format=True,
    encoding="utf-8",
    overwrite=True,
) -> None:
    """Write data to json file or file object using the custom json encoder"""
    start_timer = timer()
    if not _check_can_write(file_or_io, overwrite, verbose):
        return

    with open_file_or_io(
        file_or_io, mode="wt", encoding=encoding, create_parent=create_parent
    ) as fh:
        if indent is None and separators is None:
            separators = (",", ":")

        kwargs = dict(
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            indent=indent,
            separators=separators,
            sort_keys=sort_keys,
        )
        if custom_format:
            kwargs.update(
                dict(
                    cls=CustomJSONEncoder,
                    float_precision=float_precision,
                )
            )
        try:
            json.dump(obj, fh, **kwargs)
        except KeyboardInterrupt as e:
            if isinstance(fh, (Path, str)):
                print(f"KeyboardInterrupt, removing potentially corrupt json: {fh}")
                Path(fh).unlink()
            raise e

    if verbose:
        print(f"Wrote json file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dump_json_compressed(
    obj: Any,
    file_or_io: PathOrIO,
    compressor_name: CompressorC,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    verbose: bool = True,
    create_parent=False,
    float_precision=None,
    custom_format=True,
    encoding="utf-8",
    overwrite=True,
    **compressor_kwargs,
):
    start_timer = timer()
    if not _check_can_write(file_or_io, overwrite, verbose):
        return

    json_data = dumps_json(
        obj,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        indent=indent,
        separators=separators,
        sort_keys=sort_keys,
        float_precision=float_precision,
        custom_format=custom_format,
    )
    compress_data_to_file(
        json_data,
        file_or_io,
        compressor_name,
        encoding=encoding,
        create_parent=create_parent,
        **compressor_kwargs,
    )

    if verbose:
        print(f"Wrote json file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dumps_json(
    obj: Any,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    float_precision=None,
    custom_format=True,
) -> str:
    """Write data to json string using the custom json encoder"""
    if indent is None and separators is None:
        separators = (",", ":")

    kwargs = dict(
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        indent=indent,
        separators=separators,
        sort_keys=sort_keys,
    )
    if custom_format:
        kwargs.update(
            dict(
                cls=CustomJSONEncoder,
                float_precision=float_precision,
            )
        )
    return json.dumps(obj, **kwargs)


def dump_jsonl(
    data: Iterable[Any],
    file_or_io: PathOrIO,
    verbose: bool = True,
    create_parent=False,
    encoding="utf-8",
    overwrite=True,
) -> None:
    """Write lines of data to jsonl (list of json strings) file or file object
    using the custom json encoder"""
    start_timer = timer()
    if not _check_can_write(file_or_io, overwrite, verbose):
        return

    err_msg = f"data must be a list/sequence but is {type(data)}"
    assert not isinstance(data, str), err_msg
    assert isinstance(data, Sequence), err_msg

    with open_file_or_io(file_or_io, "wt", encoding=encoding, create_parent=create_parent) as fh:
        for d in data:
            fh.write(f"{dumps_json(d)}\n")

    if verbose:
        print(f"Wrote jsonl file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dump_jsonl_compressed(
    data: Iterable[Any],
    file_or_io: PathOrIO,
    compressor_name: CompressorC,
    verbose: bool = True,
    create_parent=False,
    encoding="utf-8",
    overwrite=True,
    **compressor_kwargs,
) -> None:
    """Write lines of data to jsonl (list of json strings) file or file object
    using the custom json encoder"""
    start_timer = timer()
    if not _check_can_write(file_or_io, overwrite, verbose):
        return

    try:
        jsonl_data = dumps_jsonl(data)
    except Exception as e:
        raise RuntimeError(f"Error dumping jsonl file {file_or_io}") from e

    compress_data_to_file(
        jsonl_data,
        file_or_io,
        compressor_name,
        encoding=encoding,
        create_parent=create_parent,
        **compressor_kwargs,
    )
    if verbose:
        print(f"Wrote jsonl file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dumps_jsonl(data: Iterable[Any]) -> str:
    sio = io.StringIO()
    dump_jsonl(data, sio, verbose=False)
    return sio.getvalue()


load_json_xz = partial(load_json_compressed, compressor_name=CompressorC.LZMA)
dump_json_xz = partial(dump_json_compressed, compressor_name=CompressorC.LZMA)
load_jsonl_xz = partial(load_jsonl_compressed, compressor_name=CompressorC.LZMA)
dump_jsonl_xz = partial(dump_jsonl_compressed, compressor_name=CompressorC.LZMA)

load_json_zst = partial(load_json_compressed, compressor_name=CompressorC.ZSTD)
dump_json_zst = partial(dump_json_compressed, compressor_name=CompressorC.ZSTD)
load_jsonl_zst = partial(load_jsonl_compressed, compressor_name=CompressorC.ZSTD)
dump_jsonl_zst = partial(dump_jsonl_compressed, compressor_name=CompressorC.ZSTD)



"""
possible improvements:
    - add more hyperparameters to the compressor wrappers, currently using mostly the defaults.
    - auto compressor name from filename
    - compressor ios similar to BytesIO that compress/decompress as needed on write/read

"""
import lzma
import zstandard
from typing import Union

from packg import Const
from packg.iotools import read_bytes_from_file_or_io, open_file_or_io


def load_xz(file, mode: str = "rt", encoding: str = "utf-8"):
    if "b" in mode:
        encoding = None
    with lzma.open(file, mode, encoding=encoding) as fh:
        data_str = fh.read()
    return data_str


def dump_xz(data, file, mode: str = "wt", encoding: str = "utf-8"):
    if "b" in mode:
        encoding = None
    with lzma.open(file, mode, encoding=encoding) as fh:
        fh.write(data)


class CompressorC(Const, str):
    NONE = "none"
    LZMA = "lzma"
    ZSTD = "zstd"
    ZSTD_SLOW = "zstd_slow"


class CompressorInterface:
    def compress(self, data: bytes) -> bytes:
        raise NotImplementedError

    def flush(self) -> bytes:
        raise NotImplementedError

    def compress_once(self, data: bytes) -> bytes:
        return b"".join([self.compress(data), self.flush()])


class DecompressorInterface:
    def decompress(self, data: bytes) -> bytes:
        raise NotImplementedError

    def flush(self) -> bytes:
        return b""

    def decompress_once(self, data: bytes) -> bytes:
        return self.decompress(data)


def decompress_file_to_bytes(file_or_io, compressor_name: str, **compressor_kwargs) -> bytes:
    data_bytes_compressed = read_bytes_from_file_or_io(file_or_io)
    return decompress_bytes_to_bytes(data_bytes_compressed, compressor_name, **compressor_kwargs)


def decompress_bytes_to_bytes(
    data_bytes_compressed: bytes, compressor_name: str, **compressor_kwargs
) -> bytes:
    decompressor = get_decompressor(compressor_name, **compressor_kwargs)
    return decompressor.decompress_once(data_bytes_compressed)


def decompress_file_to_str(
    file_or_io, compressor_name: str, encoding: str = "utf-8", **compressor_kwargs
) -> str:
    data_bytes_compressed = read_bytes_from_file_or_io(file_or_io)
    return decompress_bytes_to_str(
        data_bytes_compressed, compressor_name, encoding, **compressor_kwargs
    )


def decompress_bytes_to_str(
    data_bytes_compressed: bytes, compressor_name: str, encoding: str = "utf-8", **compressor_kwargs
) -> str:
    data_bytes = decompress_bytes_to_bytes(
        data_bytes_compressed, compressor_name, **compressor_kwargs
    )
    return data_bytes.decode(encoding)


def compress_data_to_bytes(
    data: Union[str, bytes], compressor_name: str, encoding: str = "utf-8", **compressor_kwargs
) -> bytes:
    if isinstance(data, str):
        data = data.encode(encoding)
    compressor = get_compressor(compressor_name, len(data), **compressor_kwargs)
    return compressor.compress_once(data)


def compress_data_to_file(
    data: Union[str, bytes],
    file_or_io,
    compressor_name: str,
    create_parent: bool = False,
    **compressor_kwargs,
):
    data_bytes = compress_data_to_bytes(data, compressor_name, **compressor_kwargs)
    with open_file_or_io(file_or_io, mode="wb", create_parent=create_parent) as fh:
        fh.write(data_bytes)


# noinspection PyArgumentList
def get_compressor(compressor_name: str, size: int = -1, **kwargs) -> CompressorInterface:
    """
    Args:
        compressor_name: name of the algorithm
        size: total size of the data that will be compressed.
            some compression algorithms can benefit from knowing this. default -1 = unknown
        **kwargs: parameters for the specific compressor

    Returns:
        compressor
    """
    if compressor_name == CompressorC.NONE:
        return DummyCompressor(**kwargs)
    if compressor_name == CompressorC.LZMA:
        return LzmaCompressorWrapper(**kwargs)
    if compressor_name == CompressorC.ZSTD:
        return ZstdCompressorWrapper(size=size, **kwargs)
    if compressor_name == CompressorC.ZSTD_SLOW:
        return ZstdCompressorWrapper(size=size, level=9, **kwargs)
    raise ValueError(f"Unknown compressor {compressor_name}")


# noinspection PyArgumentList
def get_decompressor(compressor_name: str, **kwargs) -> DecompressorInterface:
    """

    Args:
        compressor_name: name of the algorithm
        **kwargs: parameters for the specific decompressor

    Returns:
        decompressor
    """
    if compressor_name == CompressorC.NONE:
        return DummyDecompressor(**kwargs)
    if compressor_name == CompressorC.LZMA:
        return LzmaDecompressorWrapper(**kwargs)
    if compressor_name in (CompressorC.ZSTD, CompressorC.ZSTD_SLOW):
        return ZstdDecompressorWrapper(**kwargs)
    raise ValueError(f"Unknown compressor {compressor_name}")


class DummyCompressor(CompressorInterface):
    def compress(self, data: bytes) -> bytes:
        return data

    def flush(self) -> bytes:
        return b""


class DummyDecompressor(DecompressorInterface):
    def decompress(self, data: bytes) -> bytes:
        return data


class ZstdCompressorWrapper(CompressorInterface):
    def __init__(self, size=-1, level=3, threads=0):
        self.cctx = zstandard.ZstdCompressor(level=level, threads=threads)
        self.compressor = self.cctx.compressobj(size=size)

    def compress(self, data: bytes) -> bytes:
        return self.compressor.compress(data)

    def flush(self) -> bytes:
        return self.compressor.flush(zstandard.FLUSH_FRAME)


class ZstdDecompressorWrapper(DecompressorInterface):
    def __init__(self):
        self.cctx = zstandard.ZstdDecompressor()
        self.decompressor = self.cctx.decompressobj()

    def decompress(self, data: bytes) -> bytes:
        return self.decompressor.decompress(data)


class LzmaCompressorWrapper(CompressorInterface):
    def __init__(self):
        self.lzc = lzma.LZMACompressor()

    def compress(self, data: bytes) -> bytes:
        return self.lzc.compress(data)

    def flush(self) -> bytes:
        return self.lzc.flush()


class LzmaDecompressorWrapper(DecompressorInterface):
    def __init__(self):
        self.lzd = lzma.LZMADecompressor()

    def decompress(self, data: bytes) -> bytes:
        return self.lzd.decompress(data)



r"""
pattern:
    match(string, pos, endpos) -> re.Match | None - Determine if RE matches at beginning of string.
    fullmatch(string, pos, endpos) - same but the entire string must match
    search(string, pos, endpos) - return first match in the string
    findall(string, pos, endpos) - find all matching, nonoverlapping substrings
    finditer, same but iterator instead of list.
    split(string) - same as str.split
    sub: replace
    subn: replace and return number of replacements

match object:
    group(0) - the entire match
    group(i) - the i-th parenthesized expression
    groups() - all parenthesized expressions
    __getitem__(i) - same as group(i)
    groupdict() - named groups
        m = re.match(r"(?P<first_name>\w+) (?P<last_name>\w+)", "Malcolm Reynolds")
        m.groupdict()  # {'first_name': 'Malcolm', 'last_name': 'Reynolds'}


"""


from pathlib import Path
from pprint import pprint
from typing import Dict, Tuple, Union, BinaryIO

import urllib3
from tqdm import tqdm


def _open_file_for_download(file: Union[str, Path]) -> Tuple[Dict[str, str], BinaryIO]:
    """Open given file pointer for download and create a header to download the remainder partially.

    Args:
        file: File path.

    Returns:

    """
    file = Path(file)
    if file.is_file():
        # file exists, download remainder (nothing if download is complete)
        fsize = file.stat().st_size
        fh = file.open("ab")
    else:
        # load entire file
        fsize = 0
        fh = file.open("wb")
    headers = {"Range": f"bytes={fsize}-"}
    return headers, fh


def download_file(
    file: Union[str, Path],
    url: str,
    verbose: bool = False,
    pbar: bool = True,
    chunk_size=1024**2,
    req_header=None,
) -> int:
    """
    Download file from url.
    """
    if req_header is None:
        req_header = {}
    http = urllib3.PoolManager()
    file = Path(file)

    # support partial file downloading
    headers, fh = _open_file_for_download(file)
    headers.update(req_header)
    req: urllib3.HTTPResponse = http.request("GET", url, preload_content=False, headers=headers)
    if verbose:
        pprint(req.headers.items())

    # check how much content is left
    try:
        web_size = int(req.headers["Content-Length"])
    except KeyError:
        web_size = 1024

    # read 1MB at a time from the result
    num_bytes = 0
    if web_size == 0 or req.status == 416:
        # html code 416: range (of 0 B) not satisfiable
        if verbose:
            print(f"already downloaded.")
    else:
        if verbose or pbar:
            print(f"Downloading {web_size / 1024 ** 2:.3f} MB")
        if pbar:
            pb = tqdm(total=web_size, unit_scale=True, unit="B", unit_divisor=1024)
        while True:
            data = req.read(chunk_size)  # type: bytes
            if pbar:
                pb.update(chunk_size)
            if verbose:
                print(f"{len(data)} bytes read")
            if len(data) == 0:
                break
            fh.write(data)
            num_bytes += len(data)
        if pbar:
            pb.close()
    req.close()
    fh.close()

    if verbose:
        print(f"Downloaded {num_bytes / 1024 ** 2:.3f}MB from", url)
    return num_bytes



import os
from contextlib import contextmanager
from pathlib import Path

from packg.typext import PathType


@contextmanager
def set_working_directory(path: Path):
    """Change directory temporarily within in the context manager."""
    origin = Path(os.getcwd()).absolute()
    os.chdir(path)
    yield
    os.chdir(origin)


def get_file_size(input_file: PathType, scale: int = 0) -> float:
    """
    Args:
        input_file:
        scale: 0=B (default), 1=KB, 2=MB, 3=GB, etc.

    Returns:
        File size in specified scale
    """
    return Path(input_file).stat().st_size / 1024**scale


def get_file_size_in_mb(input_file: PathType) -> float:
    return get_file_size(input_file, scale=2)


def format_b_in_mb(size_b: int) -> str:
    return f"{size_b / 1024 ** 2:.3f}MB"


def format_b_in_gb(size_b: int) -> str:
    return f"{size_b / 1024 ** 3:.3f}GB"


def format_bytes_human_readable(size_b: int) -> str:
    scales = ["B", "KB", "MB", "GB", "TB"]
    for scale, unit in enumerate(scales):
        if size_b < 1024 ** (scale + 1) or scale == len(scales) - 1:
            return f"{size_b / 1024 ** scale:.3f}{unit}"
    raise RuntimeError("Should not reach here")



"""Base64 utilities"""


import base64


def _b64_ensure_unsafe(b64str: str):
    """Take a base64 string that is either path-/url-safe or not, and make sure
    it is unsafe, such that the decoder understands it.

    Args:
        b64str: base64 string

    Returns:
        unsafe base64 string
    """
    return b64str.replace("-", "+").replace("_", "/")


def b64_encode_from_bytes(
    bytes_in: bytes, url_safe: bool = True, strip_equals: bool = False
) -> str:
    """Encode bytes with base64.

    Args:
        bytes_in: input bytes to encode
        url_safe: replace "+" with "-" and "/" with "_" to become path/url
            safe, default True
        strip_equals: strip trailing "=" characters. those are only needed to reconstruct the
            length of the original input string.

    Returns:
        path-safe base64 string
    """
    if url_safe:
        out_str = str(base64.urlsafe_b64encode(bytes_in), "ascii")
    else:
        out_str = str(base64.b64encode(bytes_in), "ascii")
    if strip_equals:
        out_str = out_str.rstrip("=")
    return out_str


def b64_decode_to_bytes(b64str: str) -> bytes:
    """Decode a base64 string to bytes

    Args:
        b64str: base64 string

    Returns:
        decoded bytes
    """
    return base64.b64decode(_b64_ensure_unsafe(b64str))


def b64_encode_from_str(str_plain: str, encoding: str = "utf-8", url_safe: bool = True) -> str:
    """Convert string to base64 ascii string

    Args:
        str_plain: input string
        encoding: string encoding, default utf-8
        url_safe: The alphabet uses '-' instead of '+' and '_' instead of '/'.

    Returns:
        base 64 string
    """
    if url_safe:
        return str(base64.urlsafe_b64encode(bytes(str_plain, encoding)), "ascii")
    return str(base64.b64encode(bytes(str_plain, encoding)), "ascii")


def b64_decode_to_str(str_b64: str, encoding: str = "utf-8") -> str:
    """Decode base64 ascii string to original string

    Args:
        str_b64: base64 encoded string
        encoding: result string encoding, default urf-8

    Returns:
        decoded string
    """
    return str(base64.b64decode(bytes(_b64_ensure_unsafe(str_b64), "ascii")), encoding)


def get_random_b64_string(length: int = 40):
    random_bytes_as_b64 = b64_encode_from_bytes(os.urandom(length), url_safe=True)
    # at this point the string is approx 4/3 * length long, cut it down to length
    return random_bytes_as_b64[:length]



# pylint: disable=duplicate-code
# todo either fix duplicate-code or remove this todo comment
# todo move the main() example outside of this file
from __future__ import annotations

import random
import time
from multiprocessing import Process, Queue
from timeit import default_timer
from traceback import format_exception
from typing import Optional

from attr import define, field
from loguru import logger

from packg.dtime import format_seconds_adaptive
from packg.log import configure_logger
from packg.tqdmext import tqdm_max_ncols
from packg.typext import NoneType


@define
class FnMultiProcessor:
    """
    Simple multiprocessor with N producers and
        - either no output at all
        - or all output is append to an infinitely large queue and read after everything is done

    Usage:
        1. create instance
        2. loop put(arg1, arg2, ...) some tasks
        3. run()
        4. only if with_output is True: loop get() until output queue is empty
        5. close()

    Args:
        workers: number of workers (0 = foreground)
        target_fn: function to call in the worker
        ignore_errors: do not raise errors in the worker
        verbose: show progress bar
        with_output: return output from worker
        total: total number of tasks for progress bar or None if unknown
        desc: description for progress bar


    """

    workers: int
    target_fn: callable
    ignore_errors: bool = False
    verbose: bool = True
    with_output: bool = True
    total: Optional[int] = None
    desc: str = "Multiprocessing"

    worker_list: list[Process] = field(factory=list, init=False)
    q_in: Queue = field(init=False)
    q_out: Optional[Queue] = field(init=False)
    pbar: tqdm_max_ncols = field(init=False)
    start_time: int = field(init=False)
    processed: int = field(init=False)
    input_counter: int = field(init=False)

    def __attrs_post_init__(self):
        assert self.workers >= 0, f"workers must be >= 0 but is {self.workers}"
        maxsize = self.workers * 2
        self.q_in = Queue(maxsize=maxsize)
        self.q_out = Queue(maxsize=0) if self.with_output else None
        multi_fn_args = self._get_multi_fn_args()
        for _ in range(self.workers):
            w = Process(target=self._get_multi_fn(), args=multi_fn_args)
            w.start()
            self.worker_list.append(w)
        self.pbar = tqdm_max_ncols(
            total=self.total, desc=self.desc, disable=not self.verbose or self.workers == 0
        )
        self.start_time = default_timer()
        self.processed = 0
        self.input_counter = 0

    def _get_multi_fn(self):
        return multi_fn_with_output if self.with_output else multi_fn_no_output

    def _get_multi_fn_args(self):
        return self.target_fn, self.q_in, self.q_out, self.ignore_errors

    def put(self, *args):
        self.q_in.put(args)
        self.input_counter += 1
        # with W workers, the first W will be immediately accepted. 2*W will be queued.
        # so only start counting after 3*W for an accurate pbar.
        if self.input_counter > self.workers * 3:
            self.update_pbar(1)

    def update_pbar(self, amount=1):
        self.processed += amount
        sec_per_iter = (default_timer() - self.start_time) / max(self.processed, 1)
        desc_post = f" {sec_per_iter:10.2f} s/it"
        if sec_per_iter < 1:
            iter_per_sec = 1 / sec_per_iter
            desc_post = f" {iter_per_sec:10.2f} it/s"
        rem_time_fmt = ""
        if self.total is not None:
            rem_time = sec_per_iter * (self.total - self.processed)
            rem_time_fmt = format_seconds_adaptive(rem_time, "{:5.2f}{}")
        self.pbar.set_description(f"{self.desc}{desc_post} {rem_time_fmt}", refresh=False)
        self.pbar.update(amount)

    def run(self):
        # add poison pills for workers
        for _ in range(self.workers):
            self.q_in.put(None)

        if self.workers == 0:
            # no multiprocessing, queue is full, run one worker in foreground
            self.q_in.put(None)
            multi_fn_args = self._get_multi_fn_args()
            self._get_multi_fn()(*multi_fn_args, foreground=True, desc=self.desc)
        else:
            # wait for input queue to be empty
            wait_counter = 0
            delay = 0.2
            remaining = self.q_in.qsize()
            while not self.q_in.empty():
                if wait_counter % 5 == 0:
                    now_remaining = self.q_in.qsize()
                    if now_remaining < remaining:
                        delta = remaining - now_remaining
                        remaining = now_remaining
                        self.update_pbar(delta)

                wait_counter += 1
                time.sleep(delay)
            self.pbar.update(remaining)

    def get(self):
        return self.q_out.get()

    def close(self):
        # note: output queue has to be empty before workers can be joined otherwise this will hang
        logger.debug(f"Joining workers")
        for n in self.worker_list:
            n.join()
            self.update_pbar(1)
            n.terminate()
        self.pbar.close()

        self.q_in.close()
        if self.q_out is not None:
            self.q_out.close()


def multi_fn_with_output(
    fn: callable,
    in_q: Queue,
    out_q: Queue,
    ignore_errors: bool,
    foreground: bool = False,
    desc: str = "Processing",
):
    pbar = tqdm_max_ncols(total=in_q.qsize() - 1, disable=not foreground, desc=desc)
    while True:
        args = in_q.get()
        if args is None:
            break
        try:
            out = fn(*args)
        except Exception as e:
            if ignore_errors:
                logger.error(f"Error in multi_function:\n\n{''.join(format_exception(e))}")
                out_q.put(None)
                continue
            raise e
        out_q.put(out)
        pbar.update(1)
    pbar.close()


def multi_fn_no_output(
    fn: callable,
    in_q: Queue,
    out_q: NoneType,
    ignore_errors: bool,
    foreground: bool = False,
    desc: str = "Processing",
):
    assert out_q is None, "out_q must be None"
    pbar = tqdm_max_ncols(total=in_q.qsize() - 1, disable=not foreground, desc=desc)
    while True:
        args = in_q.get()
        if args is None:
            break
        try:
            fn(*args)
        except Exception as e:
            if ignore_errors:
                logger.error(f"Error in multi_function: {e}")
            else:
                raise e
        pbar.update(1)
    pbar.close()


def fn_w_out(t_x: int):
    time.sleep(random.random() * 0.2 + 0.2)
    return t_x * 2


def fn_wo_out(_t_x: int):
    time.sleep(random.random() * 0.2 + 0.2)
    # print(t_x * 3)


def main():
    configure_logger(level="INFO")
    for workers in [3, 0]:
        print(f"Workers: {workers} ******************")
        ix = list(range(10))
        proc = FnMultiProcessor(
            workers, fn_w_out, total=len(ix), desc=f"With output queue, {workers} workers"
        )
        for in_x in ix:
            proc.put(in_x)
        proc.run()
        outputs = []
        for _ in range(len(ix)):
            outputs.append(proc.get())
        proc.close()
        print(outputs)

        ix = list(range(10))
        proc = FnMultiProcessor(
            workers,
            fn_wo_out,
            with_output=False,
            total=len(ix),
            desc=f"No output queue, {workers} workers",
        )
        for in_x in ix:
            proc.put(in_x)
        proc.run()
        proc.close()


if __name__ == "__main__":
    main()



"""
Global path definitions for projects.

Resolution is as follows:
1. Load from environment variables if defined
2. Find and load from .env file in the current directory using python dotenv library
3. Use the defaults defined here

Usage in python:
    print(get_data_dir())

Usage in yaml with omegaconf:
    storage: ${oc.env:ENV_DATA_DIR}/datasetname
"""
import os
from dotenv import dotenv_values, find_dotenv
from pathlib import Path

from packg.constclass import Const

HELP_STR = """Solutions:
1. Set the environment variable in the system.
2. Create .env file in the current dir, a parent dir or user home dir and define the variable there. 
3. Change defaults in packg.paths module
"""


class EnvKeys(Const):
    ENV_DATA_DIR = "ENV_DATA_DIR"
    ENV_CACHE_DIR = "ENV_CACHE_DIR"


home = Path.home()

ENV_DEFAULTS = {
    EnvKeys.ENV_DATA_DIR: "data",  # datasets base directory, default is relative dir 'data'
    EnvKeys.ENV_CACHE_DIR: (home / ".cache").as_posix(),
}

_setup_environ_done = False


def setup_environ(
    verbose=False, load_dotenv=True, override_from_dotenv=True, dotenv_path=None, use_defaults=True
):
    global _setup_environ_done
    if _setup_environ_done and not verbose:
        return
    _setup_environ_done = True

    if load_dotenv:
        if dotenv_path is not None:
            dotenv_path = Path(dotenv_path).as_posix()
            if not Path(dotenv_path).is_file():
                raise FileNotFoundError(f"dotenv_path provided but not found: {dotenv_path}")
        else:
            dotenv_path = ""
            if verbose:
                print(f"Searching dotenv file...")

        if dotenv_path == "":
            # try to find it relative to the current dir
            dotenv_path = find_dotenv(usecwd=True)
        if dotenv_path == "":
            # try to find in home
            proposal = Path.home() / ".env"
            if proposal.is_file():
                dotenv_path = proposal.as_posix()

        if dotenv_path != "":
            values = dotenv_values(dotenv_path, verbose=verbose)
            if verbose:
                print(
                    f"Got {len(values)} from .env file, "
                    f"found it as {find_dotenv()} from {os.getcwd()}"
                )
        else:
            values = {}
            if verbose:
                print(f"Dotenv file not found.")
        for k, v in values.items():
            if override_from_dotenv or k not in os.environ:
                if verbose:
                    print(f"    From .env write: {k}={type(v).__name__} length {len(v)}")
                os.environ[k] = v

    if use_defaults:
        for env_k, v in ENV_DEFAULTS.items():
            if env_k not in os.environ:
                if verbose:
                    print(f"from packg.paths defaults write: {env_k}={v}")
                os.environ[env_k] = v


def get_from_environ(env_k: str):
    setup_environ()
    value = os.environ[env_k]
    if value == "" or value is None:
        raise ValueError(f"Environment variable {env_k} is undefined: '{value}'")
    return value


def print_all_environment_variables(print_fn=print):
    setup_environ(verbose=True)
    print_fn(f"Path definitions:")
    for env_k in EnvKeys.values():
        print_fn(f"    {env_k}={os.environ[env_k]}")


def get_data_dir() -> Path:
    return get_path_from_env(EnvKeys.ENV_DATA_DIR)


def get_cache_dir() -> Path:
    return get_path_from_env(EnvKeys.ENV_CACHE_DIR)


def get_path_from_env(env_k: str) -> Path:
    return Path(get_from_environ(env_k))



from .multiproc_fn import FnMultiProcessor, multi_fn_no_output, multi_fn_with_output
from .multiproc_producer_consumer import (
    MultiProcessorProducerConsumer,
    Producer,
    Consumer,
    SimpleProducer,
    SimpleConsumer,
)

__all__ = [
    "FnMultiProcessor",
    "multi_fn_no_output",
    "multi_fn_with_output",
    "MultiProcessorProducerConsumer",
    "Producer",
    "Consumer",
    "SimpleProducer",
    "SimpleConsumer",
]



import socket
import sys

_module = sys.modules[__name__]


def disable_socket(verbose: bool = True):
    """disable socket.socket to disable the Internet. useful in testing.
    from https://gist.github.com/hangtwenty/9200597e3be274c79896
    """
    setattr(_module, "_socket_disabled", True)
    original_socket = socket.socket

    def guarded(*args, **kwargs):
        do_raise = True
        if getattr(_module, "_socket_disabled", False) and do_raise:
            raise RuntimeError(
                "A test tried to use socket.socket without explicitly un-blocking it."
            )
        socket.socket = original_socket
        return socket.socket(*args, **kwargs)

    socket.socket = guarded
    if verbose:
        print("[!] socket.socket is now blocked. The network should be inaccessible.")


def enable_socket(verbose: bool = True):
    """re-enable socket.socket to enable the Internet."""
    setattr(_module, "_socket_disabled", False)
    if verbose:
        print("[!] socket.socket is UN-blocked, and the network can be accessed.")



"""JSON encoder implementation for jsonext.py"""
# pylint: skip-file
import json
from json.encoder import (
    c_make_encoder,  # noqa
    encode_basestring_ascii,  # noqa
    encode_basestring,  # noqa
    INFINITY,
)
from pathlib import Path

import attrs

from typedparser import NamedTupleMixin


class CustomJSONEncoder(json.JSONEncoder):
    def __init__(  # noqa
        self,
        *,
        skipkeys=False,
        ensure_ascii=False,
        check_circular=True,
        allow_nan=False,
        sort_keys=False,
        indent=None,
        separators=None,
        default=None,
        float_precision=None,
    ):
        self.skipkeys = skipkeys
        self.ensure_ascii = ensure_ascii
        self.check_circular = check_circular
        self.allow_nan = allow_nan
        self.sort_keys = sort_keys
        self.indent = indent
        self.float_precision = float_precision
        if separators is not None:
            self.item_separator, self.key_separator = separators
        # change: do not modify indent here
        if default is not None:
            self.default = default

    def default(self, o):
        # change: added more supported types.
        # note: intentionally checks for jax/torch tensors without importing these packages
        if isinstance(o, Path):
            return o.as_posix()
        full_name = f"{o.__class__.__module__}.{o.__class__.__name__}"
        # if isinstance(o, (np.int8, np.int16, np.int32, np.int64)):
        if full_name in {"numpy.int8", "numpy.int16", "numpy.int32", "numpy.int64"}:
            return int(o)
        # if isinstance(o, (np.float16, np.float32, np.float64)):
        if full_name in {"numpy.float16", "numpy.float32", "numpy.float64"}:
            return float(o)
        # if isinstance(o, np.ndarray):
        if full_name == "numpy.ndarray":
            return o.tolist()
        if hasattr(o, "detach"):  # torch
            return o.detach().cpu().numpy().tolist()
        # todo update below to use full name check once we have a jax example
        class_name = o.__class__.__name__
        if class_name.lower() == "devicearray":  # jax
            return o.tolist()
        if isinstance(o, NamedTupleMixin):
            return attrs.asdict(o)
        raise TypeError(f"Object of type {class_name} is not JSON serializable")

    def iterencode(self, o, _one_shot=False):
        # change: custom iterencode function called
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = encode_basestring_ascii
        else:
            _encoder = encode_basestring

        def floatstr(
            obj,
            allow_nan=self.allow_nan,
            _repr=float.__repr__,
            _inf=INFINITY,
            _neginf=-INFINITY,
        ):
            if obj != obj:
                text = "NaN"
            elif obj == _inf:
                text = "Infinity"
            elif obj == _neginf:
                text = "-Infinity"
            elif self.float_precision is not None:
                return f"{obj:.{self.float_precision}f}"
            else:
                return _repr(obj)  # noqa
            if not allow_nan:
                raise ValueError("Out of range float values are not JSON compliant: " + repr(obj))

            return text

        if (
            _one_shot
            and c_make_encoder is not None
            and self.indent is None
            and self.float_precision is None
        ):
            # only call the original c encoder if available and no custom iterencode logic is needed
            _iterencode = c_make_encoder(
                markers,
                self.default,
                _encoder,
                self.indent,
                self.key_separator,
                self.item_separator,
                self.sort_keys,
                self.skipkeys,
                self.allow_nan,
            )
        else:
            _iterencode = _make_custom_iterencode(
                markers,
                self.default,
                _encoder,
                self.indent,
                floatstr,
                self.key_separator,
                self.item_separator,
                self.sort_keys,
                self.skipkeys,
                _one_shot,
            )
        return _iterencode(o, 0)


def _make_custom_iterencode(
    markers,
    _default,
    _encoder,
    _indent,
    _floatstr,
    _key_separator,
    _item_separator,
    _sort_keys,
    _skipkeys,
    _one_shot,
    # HACK: hand-optimized bytecode; turn globals into locals
    ValueError=ValueError,  # noqa
    dict=dict,  # noqa
    float=float,  # noqa
    id=id,  # noqa
    int=int,  # noqa
    isinstance=isinstance,  # noqa
    list=list,  # noqa
    str=str,  # noqa
    tuple=tuple,  # noqa
    _intstr=int.__repr__,
    indent_list: bool = False,
):
    if _indent is not None and not isinstance(_indent, str):
        _indent = " " * _indent

    # change: no spaces at the end of lines
    _item_separator_eol = _item_separator.rstrip()

    def _iterencode_list(lst, _current_indent_level):
        if not lst:
            yield "[]"
            return
        if markers is not None:
            markerid = id(lst)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = lst
        buf = "["
        # # change: use indent_list flag, use eol separator
        if _indent is not None and indent_list:
            _current_indent_level += 1
            newline_indent = "\n" + _indent * _current_indent_level
            separator = _item_separator_eol + newline_indent
            buf += newline_indent
        else:
            newline_indent = None
            separator = _item_separator
        first = True
        for value in lst:
            if first:
                first = False
            else:
                buf = separator
            if isinstance(value, str):
                yield buf + _encoder(value)
            elif value is None:
                yield buf + "null"
            elif value is True:
                yield buf + "true"
            elif value is False:
                yield buf + "false"
            elif isinstance(value, int):
                # Subclasses of int/float may override __repr__, but we still
                # want to encode them as integers/floats in JSON. One example
                # within the standard library is IntEnum.
                yield buf + _intstr(value)  # noqa
            elif isinstance(value, float):
                # see comment above for int
                yield buf + _floatstr(value)
            else:
                yield buf
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield "\n" + _indent * _current_indent_level
        yield "]"
        if markers is not None:
            del markers[markerid]  # noqa

    def _iterencode_dict(dct, _current_indent_level):
        if not dct:
            yield "{}"
            return
        if markers is not None:
            markerid = id(dct)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = dct
        yield "{"
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = "\n" + _indent * _current_indent_level
            item_separator = _item_separator_eol + newline_indent
            yield newline_indent
        else:
            newline_indent = None
            item_separator = _item_separator
        first = True
        if _sort_keys:
            items = sorted(dct.items())
        else:
            items = dct.items()
        for key, value in items:
            if isinstance(key, str):
                pass
            # JavaScript is weakly typed for these, so it makes sense to
            # also allow them.  Many encoders seem to do something like this.
            elif isinstance(key, float):
                # see comment for int/float in _make_iterencode
                key = _floatstr(key)
            elif key is True:
                key = "true"
            elif key is False:
                key = "false"
            elif key is None:
                key = "null"
            elif isinstance(key, int):
                # see comment for int/float in _make_iterencode
                key = _intstr(key)  # noqa
            elif _skipkeys:
                continue
            else:
                raise TypeError(
                    f"keys must be str, int, float, bool or None, " f"not {key.__class__.__name__}"
                )
            if first:
                first = False
            else:
                yield item_separator
            yield _encoder(key)
            yield _key_separator
            if isinstance(value, str):
                yield _encoder(value)
            elif value is None:
                yield "null"
            elif value is True:
                yield "true"
            elif value is False:
                yield "false"
            elif isinstance(value, int):
                # see comment for int/float in _make_iterencode
                yield _intstr(value)  # noqa
            elif isinstance(value, float):
                # see comment for int/float in _make_iterencode
                yield _floatstr(value)
            else:
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield "\n" + _indent * _current_indent_level
        yield "}"
        if markers is not None:
            del markers[markerid]  # noqa

    def _iterencode(o, _current_indent_level):
        if isinstance(o, str):
            yield _encoder(o)
        elif o is None:
            yield "null"
        elif o is True:
            yield "true"
        elif o is False:
            yield "false"
        elif isinstance(o, int):
            # see comment for int/float in _make_iterencode
            yield _intstr(o)  # noqa
        elif isinstance(o, float):
            # see comment for int/float in _make_iterencode
            yield _floatstr(o)
        elif isinstance(o, (list, tuple)):
            yield from _iterencode_list(o, _current_indent_level)
        elif isinstance(o, dict):
            yield from _iterencode_dict(o, _current_indent_level)
        else:
            if markers is not None:
                markerid = id(o)
                if markerid in markers:
                    raise ValueError("Circular reference detected")
                markers[markerid] = o
            o = _default(o)
            yield from _iterencode(o, _current_indent_level)
            if markers is not None:
                del markers[markerid]  # noqa

    return _iterencode



import time
from typing import Optional

from attrs import define

from typedparser import add_argument


def connect_to_pycharm_debug_server(host: Optional[str] = None, port: Optional[int] = None) -> None:
    """
    Connect to a running pycharm debug server.

    To setup in PyCharm See Debug -> Edit Configurations -> Add New Configuration ->
        Python Debug Server -> Configure port, install packages, etc. -> Run

    Then run this function at the start of your code to connect it to the server.

    Args:
        host: host or None for skipping the connection process
        port: port or None for default port 12345
    """
    if host is None:
        return
    import pydevd_pycharm  # noqa  # pylint: disable=import-error

    if port is None:
        port = 12345
    while True:
        try:
            pydevd_pycharm.settrace(
                host, port=port, stdoutToServer=True, stderrToServer=True, suspend=False
            )
            break
        except ConnectionRefusedError:
            print(f"Debug server connection refused: {host}:{port}. Retrying...")
            time.sleep(2)

@define
class PyCharmDebugArgs:
    trace: str | None = add_argument(
        type=str, help="Connect debug server on this host.", default=None
    )
    trace_port: int = add_argument(type=int, default=33553, help="Target debugging server port")


"""Type extensions"""
from io import IOBase
from pathlib import Path
from typing import Union, Any

from importlib_resources.abc import Traversable

TensorType = Any
PathType = Union[str, Path, Traversable]
OptionalPathType = Union[str, Path, Traversable, None]
PathOrIO = Union[PathType, IOBase]

# Subscripted generics cannot be used with class and instance check
# i.e. to use isinstance(obj, the_type) they need to be defined as a tuple
PathTypeCls = (str, Path, Traversable)
PathOrIOCls = (str, Path, Traversable, IOBase)

try:
    from types import NoneType  # pylint: disable=unused-import
except ImportError:
    NoneType = type(None)



from collections import defaultdict
from copy import deepcopy
from typing import List, Dict, Tuple


def create_unique_abbreviations(input_strings: List[str], seps=("_", ".")) -> Dict[str, str]:
    """

    Old version, splits input strings at "_" and "." and takes first letter of each word.

    lg         run.list_gpus
    pcbs       run.packaging.create_build_scripts

    Args:
        input_strings: list of strings to abbreviate
        seps: separator for splitting the input strings

    Returns:
        dict {input_string: abbreviation}
    """
    abbreviations = {}
    for name in input_strings:
        name_for_abbr = name
        for sep in seps[1:]:
            name_for_abbr = name_for_abbr.replace(sep, seps[0])
        words = name_for_abbr.split(seps[0])
        abbreviation = ""
        for word in words:
            abbreviation += word[0]
        if abbreviation in abbreviations:
            # enforce uniqueness
            i = 1
            while True:
                new_abbreviation = abbreviation + str(i)
                if new_abbreviation not in abbreviations:
                    abbreviation = new_abbreviation
                    break
                i += 1
        abbreviations[abbreviation] = name
    return abbreviations


def create_nested_abbreviations(input_strings: List[str], sep_dir=".") -> Dict[str, str]:
    """
    New version, keep the directory nesting and use minimal amount of letters

    l          run.list_gpus
    p.c        run.packaging.create_build_scripts

    Args:
        input_strings: list of strings to abbreviate
        sep_dir: separator for splitting the input strings into directories

    Returns:
        dict {input_string: abbreviation}
    """
    assert len(set(input_strings)) == len(input_strings), "Input strings must be unique"

    # split inputs and sort them into a nested dictionary, leaf nodes are empty dicts
    dct = {}
    for name in input_strings:
        parts = name.split(sep_dir)
        c_dct = dct
        for part in parts:
            if part not in c_dct:
                c_dct[part] = {}
            c_dct = c_dct[part]

    def _shorten_keys(keys: List[str]) -> List[Tuple[str, str]]:
        """For a given list of flat keys, find the shortest non-common prefix for each key.
        Returns a list of tuples (short_key, long_key)
        """
        if len(keys) == 0:
            return []
        longest_len = max(len(key) for key in keys)
        remaining_keys = deepcopy(keys)
        done_keys = []
        for k in range(1, longest_len + 1):
            # for a cutoff length k, sort the cut keys into buckets
            cut_keys = defaultdict(list)
            for key in remaining_keys:
                short_key = key[:k]
                cut_keys[short_key].append(key)

            # buckets with size 1 are done, for other buckets the cutoff needs to increase
            remaining_keys = []
            for short_key, t_keys in cut_keys.items():
                if len(t_keys) == 1:
                    done_keys.append((short_key, t_keys[0]))
                else:
                    remaining_keys.extend(t_keys)

            if len(remaining_keys) == 0:
                break
        else:
            raise ValueError("Could not find abbreviations")  # should not happen ever
        return done_keys

    def _recursive_shortcuts(dct_in, stem_short="", stem_long=""):
        """
        Given the nested dictionary, find the shortest non-common prefix for each nested key.
        Accumulate the short and long stems from the recursive calls.
        """
        if len(dct_in) == 0:
            return []
        keys = list(dct_in.keys())

        # shorten leafs and non-leafs separately as they dont need to be unique between each other
        key_is_leaf = {key: len(dct_in[key]) == 0 for key in keys}
        leaf_keys = [key for key, is_leaf in key_is_leaf.items() if is_leaf]
        non_leaf_keys = [key for key, is_leaf in key_is_leaf.items() if not is_leaf]
        short_leaf_keys = _shorten_keys(leaf_keys)
        short_non_leaf_keys = _shorten_keys(non_leaf_keys)

        return_strs = []
        for short_key, long_key in short_leaf_keys + short_non_leaf_keys:
            sub_stem_short = f"{stem_short}.{short_key}"
            sub_stem_long = f"{stem_long}.{long_key}"
            if key_is_leaf[long_key]:
                # leaf node, return the short and long total stems
                return_strs.append((sub_stem_short, sub_stem_long))
                continue
            # non-leaf node, recurse
            sub_strs = _recursive_shortcuts(dct_in[long_key], sub_stem_short, sub_stem_long)
            return_strs.extend(sub_strs)
        return return_strs

    # finally cut the first "." and return the dict
    return {s[1:]: l[1:] for s, l in _recursive_shortcuts(dct)}



from pathlib import Path
from typing import Dict

from packg.iotools.file_reader import yield_lines_from_object
from packg.iotools.misc import set_working_directory
from packg.system import systemcall_with_assert
from packg.typext import PathType


def get_gitadd_dryrun_output(base_dir):
    with set_working_directory(base_dir):
        out, _err, _retcode = systemcall_with_assert("git add --dry-run .")
    return out


def parse_gitadd_dryrun_output(
    output: str, base_dir: PathType, return_relative_paths: bool = True
) -> Dict[str, str]:
    """
    Parse the output of git add --dry-run . into a dict of files and their status.

    Notes:
        Alternative way to do this would be git status: https://git-scm.com/docs/git-status#_output
    """
    files = {}
    for line in yield_lines_from_object(output.splitlines()):
        line_split = line.split(" ")
        status = line_split[0].strip()
        file = " ".join(line_split[1:]).strip()[1:-1]
        file_pth = Path(base_dir) / file
        if status == "add":
            assert (
                file_pth.exists()
            ), f"Could not find path {file_pth} given line {line} and basedir {base_dir}"
        elif status == "remove":
            assert (
                not file_pth.exists()
            ), f"Found path {file_pth} given line {line} and basedir {base_dir}"
        else:
            raise ValueError(f"Unknown status {status}")

        file_return = file_pth.as_posix()
        if return_relative_paths:
            file_return = Path(file).as_posix()
        files[file_return] = status
    return files


def check_for_stages(base_dir):
    """
    Check if there are staged files in the given directory.

    Returns:
        Non-empty string if there are staged files, "" otherwise
    """
    with set_working_directory(base_dir):
        out, _err, _retcode = systemcall_with_assert("git diff --staged --name-only")
    return out.strip()


def check_for_unpushed_commits(base_dir):
    """
    Check if there are unpushed commits in the given directory.

    Returns:
        Non-empty string if there are staged files, "" otherwise
    """
    with set_working_directory(base_dir):
        out, _err, _retcode = systemcall_with_assert("git log --branches --not --remotes")
    return out.strip()


def get_gitstatusz_output(base_dir):
    with set_working_directory(base_dir):
        out, _err, _retcode = systemcall_with_assert("git status -z")
    return out


def parse_gitstatusz_output(output):
    """
    Currently unused!
    Parse the output of git status -z into a dict of files and their status.

    output:
        {path: [x, y, source_path or None]}
        source_path will be filed if path was renamed or copied to source_path

    """
    files = {}
    lines = output.split("\x00")
    c_pos = 0
    while True:
        try:
            line = lines[c_pos]
        except IndexError:
            break
        if line.strip() == "":
            c_pos += 1
            continue
        print(f"{c_pos}: {line}")
        assert line[2] == " ", f"Expected line to start with 'XY ' but got {line}"
        x, y, path = line[0], line[1], line[3:]
        other_path = None
        if x in "RC" or y in "RC":
            # for rename and copy the next line contains the other path without status
            other_path = lines[c_pos + 1]
            c_pos += 1
        files[path] = [x, y, other_path]
        c_pos += 1
    return files



"""
Utilities for calendar dates and times.

Note: If more timezones than system and UTC are required, use pytz package
"""
import datetime
from typing import Optional


def _datetime_now(tz=None):  # this function allows to mock datetime.now() in tests
    return datetime.datetime.now(tz=tz)


def get_system_timezone() -> datetime.tzinfo:
    return datetime.datetime.now().astimezone().tzinfo


def get_utc_timezone() -> datetime.tzinfo:
    # print(type(datetime.timezone(datetime.timedelta(minutes=0))))
    return datetime.timezone(datetime.timedelta(minutes=0))


def get_timestamp_for_filename(dtime: Optional[datetime.datetime] = None, tz=None) -> str:
    """
    Convert datetime to timestamp for filenames.

    Args:
        dtime: Optional datetime object, will use now() if not given.
        tz: Optional timezone object for using now(), will use system timezone if not given.

    Returns:
        string like 1970_12_31_23_59_59
    """
    if dtime is None:
        dtime = _datetime_now(tz=tz)
    ts = dtime.strftime("%Y_%m_%d_%H_%M_%S")
    return ts


def format_seconds_adaptive(seconds: float, format_str="{:.1f}{}"):
    """
    Format seconds to adaptive time format depending on the size of the input.

    Args:
        seconds: number of seconds
        format_str: format string for number and time unit

    Returns:
        string like 1.2s, 1.2min, 1.2h, 1.2d
    """
    abs_seconds = abs(seconds)
    if abs_seconds < 60:
        number = seconds
        unit = "s"
    elif abs_seconds < 3600:
        number = seconds / 60
        unit = "min"
    elif abs_seconds < 3600 * 24:
        number = seconds / 3600
        unit = "h"
    elif abs_seconds < 3600 * 24 * 7:
        number = seconds / (3600 * 24)
        unit = "d"
    elif abs_seconds < 3600 * 24 * 365:
        number = seconds / (3600 * 24 * 7)
        unit = "w"
    else:
        number = seconds / (3600 * 24 * 365)
        unit = "y"
    return format_str.format(number, unit)


def format_timestamp(
    timestamp: Optional[int] = None,
    format_str="%Y-%m-%d %H:%M:%S",
    tz=None,
) -> str:
    """
    Format a unix timestamp as date.

    Args:
        timestamp: unix timestamp
        format_str: output format
        tz: Optional timezone object for using now(), will use system timezone if not given.

    Returns:
        date string
    """
    if timestamp is None:
        dt_object = _datetime_now(tz=tz)
    else:
        dt_object = datetime.datetime.fromtimestamp(timestamp, tz=tz)
    return dt_object.strftime(format_str)



"""
Utilities for python packages.
"""
from __future__ import annotations
import importlib
import re
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional

import requests
from loguru import logger

from packg.iotools import sort_file_paths_with_dirs_separated
from packg.misc import format_exception
from packg.strings import create_nested_abbreviations
from packg.testing.import_from_source import recurse_modules


def run_package(main_file, run_dir="cli", recursive=True, run_dir_only=True, abbreviations=True):
    """
    todo remove this. it is too slow and ineffective. solve the problem in bash instead.

    Create a command line interface for a package given a directory of scripts.

    Check the run directory, list scripts, create shortcuts, run if given shortcut.

    Args:
        main_file: __file__ attribute of __main__.py e.g.
            /path/to/lmbtools/python/lmbtools/__main__.py
        run_dir: submodule with run scripts
        recursive: if True, recursively search for scripts
        run_dir_only: if True, only enable running scripts in the run_dir
        abbreviations: if True, create abbreviations for the scripts
    """
    main_file = Path(main_file)
    print(f"main_file: {main_file}")
    package_dir = main_file.parent
    package_name = package_dir.name
    package_scripts_dir = package_dir / run_dir
    glob_str = "**/*.py" if recursive else "*.py"

    package_scripts = [
        f.relative_to(package_scripts_dir)
        for f in package_scripts_dir.glob(glob_str)
        if not f.name.startswith("__")
    ]
    sorted_scripts = sort_file_paths_with_dirs_separated(package_scripts, dirs_first=False)
    package_scripts = [f.as_posix()[:-3].replace("/", ".") for f in sorted_scripts]

    abbrevs = None
    if abbreviations:
        abbrevs = create_nested_abbreviations(package_scripts)

    args = sys.argv[1:]
    if len(args) > 0:
        abbrev_arg = args[0]
        if abbrevs is not None and abbrev_arg in abbrevs:
            script = abbrevs[abbrev_arg]
            target = f"{package_name}.{run_dir}.{script}"
            run_script(target, args)
            return

        target = f"{package_name}.{abbrev_arg}"
        if run_dir_only:
            target = f"{package_name}.{run_dir}.{abbrev_arg}"
        try:
            importlib.import_module(target)
            found = True
        except ModuleNotFoundError as e:
            # must figure out whether the error is due to the module not existing
            # or due to some import failing.
            if f"'{target}'" in str(e):
                # No module named (the module we are trying to run) -> module not found
                found = False
            else:
                # No module named (some other module) -> module found, but has import error
                # run it to trigger the error
                found = True
        if found:
            run_script(target, args)
            return
        print(f"Error, script not found: {abbrev_arg} and {target}")

    # no arguments, error, show help
    if abbreviations:
        print(f"Usage option 1: {package_name} shortcut [args]")
        print(f"Usage option 2: {package_name} path.to.module [args]")
        print()
        print(f"shortcut   script")
        for abbrev, f in abbrevs.items():
            if run_dir_only:
                print(f"{abbrev:10s} {f}")
            else:
                print(f"{abbrev:10s} {run_dir}.{f}")
        return

    # show help but without abbreviations
    print(f"Usage: {package_name} path.to.module [args]")
    print()
    print(f"Available scripts:")
    for f in package_scripts:
        if run_dir_only:
            print(f"{f}")
        else:
            print(f"{run_dir}.{f}")


def run_script(target, args):
    cmd = ["python", "-m", target, *args[1:]]
    print(f"$ {' '.join(cmd)}")

    subprocess.run(cmd)

    # # run by importing main() function
    # print(f"Running {target}")
    # sys.argv = [sys.argv[0]] + args[1:]
    # module = importlib.import_module(target)
    # main = getattr(module, "main")
    # main()


def _recurse_modules_remove_root_package(package: str, packages_only: bool = False):
    for candidate in recurse_modules(package, packages_only=packages_only):
        if candidate == package:
            continue
        candidate = candidate.removeprefix(f"{package}.")
        if candidate == "__main__":
            continue
        yield candidate


def get_modules_for_autocomplete(package: str, run_dir: Optional[str] = None):
    packages_only = set(_recurse_modules_remove_root_package(package, packages_only=True))
    all_modules = set(_recurse_modules_remove_root_package(package))
    output_modules = []

    for m in sorted(all_modules):
        if m in packages_only:
            mlog = f"{package}.{m}"
            if f"{m}.__main__" not in all_modules:
                logger.debug(f"SKIP package: {mlog}")
                continue
            logger.debug(f"ADD package (has __main__): {mlog}")
        if m.endswith("__main__"):
            logger.debug(f"SKIP __main__ file: {mlog}")
            continue
        output_modules.append(m)
    output_modules.sort()

    if run_dir is not None:
        new_output_modules = []
        for output_module in output_modules:
            if output_module.startswith(run_dir):
                new_output_modules.append(output_module)
        output_modules = new_output_modules
    return output_modules


def create_bash_autocomplete_script(
    package: str, function_name: str = None, command_name: str = None, run_dir: Optional[str] = None
) -> str:
    """
    Create a bash script for autocompletion of a python package.

    Workflow where this is useful:
    - src/packg/__main__.py calls run_package, so `python -m packg run.script` has the same effect
    as `python -m packg.run.script`
    - pyproject.toml defines script [project.scripts] packg = "packg.__main__:main"
    so now `packg run.script` has the same effect.
    - this function now can be used to create src/packg/autocomplete.sh which, when sourced,
    provides autocompletion for the `packg` command.

    Args:
        package: package to create the autocomplete for
        function_name: default _{package}
        command_name: default {package}
        run_dir: only create autocompletion for this directory

    Returns:

    """
    output_modules = get_modules_for_autocomplete(package, run_dir=run_dir)
    if function_name is None:
        function_name = f"_{package}"
    if command_name is None:
        command_name = package
    ob, cb = "{", "}"
    autocomplete_script = f"""
{function_name}() {ob}
    local cur prev opts
    _init_completion || return
    # complete first argument with script
    if [ $COMP_CWORD -eq 1 ]; then
        opts="{' '.join(output_modules)}"
        COMPREPLY=( $( compgen -W "${ob}opts{cb}" -- "${ob}cur{cb}") )
        return 0
    fi
    # # otherwise complete with filesystem
    # COMPREPLY=( $(compgen -f -- "${ob}cur{cb}") )
    _filedir
    return 0
{cb}

complete -F {function_name} {command_name}
"""
    return autocomplete_script


def create_new_bash_autocomplete_script(
    packages: list[str],
    command_name: str = "py",
    function_name: Optional[str] = None,
    run_dir: Optional[str] = None,
) -> str:
    """
    New version: allinone e.g. alias py and then complete for all installed packages.

    Args:
        packages: packages to create the autocomplete for
        function_name: default _{package}
        command_name: default {package}
        run_dir: only create autocompletion for this directory

    Returns:

    """
    all_output_modules = []
    for package in packages:
        output_modules = get_modules_for_autocomplete(package, run_dir=run_dir)
        output_modules = [f"{package}.{m}" for m in output_modules]
        logger.debug(f"package={package} output_modules={output_modules}")
        if len(output_modules) == 0:
            logger.warning(f"Nothing found for package {package} in {run_dir}, is it installed?")
        all_output_modules.extend(output_modules)
    if function_name is None:
        function_name = f"_{command_name}"

    ob, cb = "{", "}"
    autocomplete_script = f"""
{function_name}() {ob}
    local cur prev opts
    _init_completion || return
    # complete second argument with script, iff first argument is -m
    if [[ $COMP_CWORD -eq 2 && $prev == "-m" ]]; then
        opts="{' '.join(all_output_modules)}"
        COMPREPLY=( $( compgen -W "${ob}opts{cb}" -- "${ob}cur{cb}") )
        return 0
    fi
    # # otherwise complete with filesystem
    # COMPREPLY=( $(compgen -f -- "${ob}cur{cb}") )
    _filedir
    return 0
{cb}

complete -F {function_name} {command_name}
"""
    return autocomplete_script


FILEDIR_AUTOCOMPLETE = r"""
_filedir()  # source: ubuntu 2004 /usr/share/bash-completion/bash_completion
{
    local IFS=$'\n'
    _tilde "$cur" || return
    local -a toks
    local reset
    if [[ "$1" == -d ]]; then
        reset=$(shopt -po noglob); set -o noglob
        toks=( $(compgen -d -- "$cur") )
        IFS=' '; $reset; IFS=$'\n'
    else
        local quoted
        _quote_readline_by_ref "$cur" quoted
        # Munge xspec to contain uppercase version too
        # http://thread.gmane.org/gmane.comp.shells.bash.bugs/15294/focus=15306
        local xspec=${1:+"!*.@($1|${1^^})"} plusdirs=()
        # Use plusdirs to get dir completions if we have a xspec; if we don't,
        # there's no need, dirs come along with other completions. Don't use
        # plusdirs quite yet if fallback is in use though, in order to not ruin
        # the fallback condition with the "plus" dirs.
        local opts=( -f -X "$xspec" )
        [[ $xspec ]] && plusdirs=(-o plusdirs)
        [[ ${COMP_FILEDIR_FALLBACK-} ]] || opts+=( "${plusdirs[@]}" )
        reset=$(shopt -po noglob); set -o noglob
        toks+=( $(compgen "${opts[@]}" -- $quoted) )
        IFS=' '; $reset; IFS=$'\n'
        # Try without filter if it failed to produce anything and configured to
        [[ -n ${COMP_FILEDIR_FALLBACK:-} && -n "$1" && ${#toks[@]} -lt 1 ]] && {
            reset=$(shopt -po noglob); set -o noglob
            toks+=( $(compgen -f "${plusdirs[@]}" -- $quoted) )
            IFS=' '; $reset; IFS=$'\n'
        }
    fi
    if [[ ${#toks[@]} -ne 0 ]]; then
        # 2>/dev/null for direct invocation, e.g. in the _filedir unit test
        compopt -o filenames 2>/dev/null
        COMPREPLY+=( "${toks[@]}" )
    fi
} # _filedir()

"""


def _get_raw_shields_io_output(package: str):
    # todo wrap this request code into a separate function in packg.web
    url = f"https://img.shields.io/pypi/v/{package}"
    logger.info(f"Finding pypi version via {url}")
    counter = 0
    while True:
        try:
            response = requests.get(url, timeout=10)
            break
        except Exception as e:
            logger.warning(f"Error requesting {url}: {format_exception(e)}")
        counter += 1
        time.sleep(1)
    return response.text


_RE_TEXT = re.compile(r"<text.*?>.*?</text>")
_RE_TEXT_VERSION = re.compile(r"<text.*?>v([0-9]+.[0-9]+.[0-9]+)</text>")


def find_pypi_package_version(package: str) -> Optional[str]:
    """
    Find the latest version of a package on pypi.

    pip search was removed. An alternative is to query the shields.io service and parse the
    output svg.
    https://img.shields.io/pypi/v/numpy

    Args:
        package: package name

    Returns:
        version string

    """
    raw = _get_raw_shields_io_output(package)
    # find all occurrences of <text>...</text>
    matches = _RE_TEXT.findall(raw)
    # the last one is the displayed version
    version = matches[-1]
    # find the version number
    match = _RE_TEXT_VERSION.search(version)
    if match:
        version = match.group(1)
        return version
    logger.error(f"Could not find version for {package} on shields.io")
    return None


def main():
    print(_get_raw_shields_io_output("numpy"))


if __name__ == "__main__":
    main()



"""
Some reusable fixtures and example setups for writing the tests.
"""

import os
import shutil
import tempfile
from pathlib import Path

import pytest

from packg.iotools.misc import set_working_directory


@pytest.fixture(scope="session", autouse=True)
def session_tmp_path():
    tmp_dir = tempfile.mkdtemp()
    yield Path(tmp_dir)
    shutil.rmtree(tmp_dir, ignore_errors=True)


@pytest.fixture(scope="session", autouse=True)
def git_example_fixture(session_tmp_path):
    """
    Setup a git example directory git_dir s.t. `git status --porcelain` returns
         M a
         D b
        R  c -> c2
        A  f
        ?? e

    Not using git -C git_dir because that makes adding files to the repo more difficult
    """
    git_dir = session_tmp_path / "git_example"
    setup_git_example(git_dir)
    yield git_dir


def setup_git_example(git_dir):
    os.makedirs(git_dir, exist_ok=True)

    def _write_text(p, t):
        Path(p).write_text(t, encoding="utf-8")

    def verbose_os_system(cmd):
        print("$", cmd)
        os.system(cmd)

    with set_working_directory(git_dir):
        verbose_os_system("git init")
        # author config is necessary e.g. on test job runners, otherwise git commit fails
        verbose_os_system('git config user.email "test@te.st"')
        verbose_os_system('git config user.name "Te St"')
        _write_text("a", "a")
        _write_text("b", "b")
        _write_text("c", "c")
        _write_text("d", "d")
        _write_text("g", "g")
        verbose_os_system("git add .")
        verbose_os_system("git commit -q -m initcommit")
        verbose_os_system("git log --oneline")
        _write_text("a", "a2")
        os.remove("b")
        os.rename("c", "c2")
        _write_text("e", "e")
        _write_text("f", "f")
        _write_text("g", "g2")
        verbose_os_system("git add c c2 f g")

        # verbose_os_system("git status")  # human readable
        # verbose_os_system("git status --porcelain")  # stable, short
        # out, _err, _retcode= systemcall("git status -z")  # machine readable
        # print("OUT", out.replace("\0", "\n"))
        # verbose_os_system("git add --dry-run .")  # add 'a'\nremove 'b'\n ...



from .abbreviations import create_nested_abbreviations, create_unique_abbreviations
from .base64tools import (
    b64_encode_from_str,
    b64_encode_from_bytes,
    b64_decode_to_str,
    b64_decode_to_bytes,
    get_random_b64_string,
)
from .hasher import hash_object
from .quote_urlparse import quote_with_urlparse, unquote_with_urlparse



"""
Utilities to index and sort file trees.
"""

from __future__ import annotations

import itertools
import os
import re
from operator import itemgetter
from pathlib import Path
from typing import Iterator, Union, Optional

import natsort
from attr import define
from tqdm import tqdm

from packg.iotools.pathspec_matcher import (
    PathSpecArgs,
    make_pathspecs,
    expand_pathspec_args,
    SPECLISTTYPE,
    apply_pathspecs,
)
from packg.typext import PathType
from typedparser import NamedTupleMixin


def regex_glob(
    base_path: PathType,
    regex_pattern: Union[re.Pattern, str],
    glob_pattern: str = "**/*",
    match_filename_only: bool = False,
    match_inverse: bool = False,
    ignore_directories: bool = False,
    return_relative_paths: bool = False,
    return_as_posix_str: bool = False,
) -> Iterator[Union[PathType, str]]:
    """Glob with regex filter

    Args:
        base_path: base path
        regex_pattern: regex pattern to filter
        glob_pattern: glob pattern to filter, potentially increase performance by pre-filtering
        match_filename_only: match only the filename, not the full path
        match_inverse: return files that do not match the regex
        ignore_directories: ignore directories in output results
        return_relative_paths: return relative paths instead of absolute paths
        return_as_posix_str: return as posix string instead of Path object

    Returns:
        list of paths
    """
    # print(f"Got pattern {regex_pattern}")
    if isinstance(regex_pattern, str):
        # print(f"Compiling {regex_pattern}")
        regex_pattern = re.compile(regex_pattern)
        # print(f"Got {regex_pattern}")
    glob_results = list(Path(base_path).glob(glob_pattern))
    # print("got", len(glob_results),"results for", base_path, glob_pattern,":\n", glob_results)
    for pth in glob_results:
        if match_filename_only:
            str_to_match = pth.name
        else:
            str_to_match = pth.as_posix()
        re_matches_bool = bool(regex_pattern.search(str_to_match))
        # print("match", re_matches_bool, "for", str_to_match, "with", regex_pattern)
        if match_inverse:
            re_matches_bool = not re_matches_bool
        if re_matches_bool:
            if ignore_directories and pth.is_dir():
                continue
            if return_relative_paths:
                pth = pth.relative_to(base_path)
            if return_as_posix_str:
                pth = pth.as_posix()
            yield pth


def sort_file_paths_with_dirs_separated(
    file_paths: list[PathType], natsorted: bool = False, dirs_first: bool = True
) -> list[Path]:
    """
    Sort a list of file paths, separating files inside subdirectories from files in the root.

    Only works for file paths (not directory paths) - input "dir/" will be treated like "dir"
    since in pathlib, Path("dir/") and Path("dir") are the same thing.

    Args:
        file_paths:
        natsorted: natural sort (image1, image2, image10) instead of (image1, image10, image2)
        dirs_first: if True, sort directories before files, otherwise sort files before dirs

    Returns:
        sorted list of paths
    """
    paths = [Path(p) for p in file_paths]
    sort_index_dir = 0 if dirs_first else 2
    # split path into its parts, then create a list of (sort_index, part) tuple for the path
    key_paths = [
        [(sort_index_dir, part) for part in path.parts[:-1]] + [(1, path.name)] for path in paths
    ]
    sort_fn = natsort.natsorted if natsorted else sorted
    sorted_tuples = sort_fn(zip(paths, key_paths), key=itemgetter(1))
    sorted_paths = [p[0] for p in sorted_tuples]
    return sorted_paths


@define
class FileProperties(NamedTupleMixin):
    size: int
    mtime: float


_total_counter, _ignored_dirs_counter, _ignored_files_counter = 0, 0, 0
_pbar: tqdm = None
_spinner = itertools.cycle("|/-\\")
_max_print = 40  # cut filenames when output is verbose
_count_every = 500  # how often to update the spinner


def make_index(
    base_root: PathType,
    recursive: bool = True,
    verbose: bool = True,
    reverse: bool = False,
    show_file_if_verbose: bool = False,
    pathspec_args: Optional[PathSpecArgs] = None,
    follow_symlinks: bool = False,
) -> dict[str, FileProperties]:
    """Create file index dictionary of some path

    Args:
        base_root: input path
        recursive: recurse into subdirs
        verbose: verbosity (default True: print progress)
        reverse: output files in reverse order
        show_file_if_verbose: show file names in verbose mode (default False: only show file sizes)
        pathspec_args: optional pathspec arguments to filter files
        follow_symlinks: follow symlinks (default False) both by recursing into symlinked dirs and
            by indexing symlinked files.

    Returns:
        file dict {filename str : (file_size int, time_last_modified float) }
    """
    base_root = Path(base_root).resolve().absolute()
    global _total_counter, _pbar, _ignored_dirs_counter, _ignored_files_counter
    _total_counter = 0
    _pbar = tqdm(total=0, disable=not verbose, desc="Indexing files")
    specs = []
    if pathspec_args is not None:
        specs = make_pathspecs(**expand_pathspec_args(pathspec_args))
    file_list = _recursive_index(
        base_root,
        base_root,
        0,
        recursive=recursive,
        verbose=verbose,
        reverse=reverse,
        show_file_if_verbose=show_file_if_verbose,
        specs=specs,
        follow_symlinks=follow_symlinks,
    )
    file_dict = {}
    for filename, size, modtime in file_list:
        file_dict[filename] = FileProperties(size, modtime)
    assert len(file_dict) == len(file_list)
    _pbar.set_description(
        f"Indexed {len(file_dict)} files. Ignored {_ignored_dirs_counter} dirs and "
        f"{_ignored_files_counter} files."
    )
    _pbar.clear()
    _pbar.close()
    return file_dict


def _recursive_index(
    root: Path,
    base_root: Path,
    depth: int,
    recursive: bool = True,
    verbose: bool = True,
    reverse: bool = False,
    show_file_if_verbose: bool = False,
    specs: Optional[SPECLISTTYPE, None] = None,
    follow_symlinks: bool = False,
) -> list[tuple[str, int, float]]:
    """Recursive helper function for make_index

    Args:
        root: current root
        base_root: total root where the indexing started
        depth: current depth

    Returns:
        list of tuples (filename str, file_size_bytes int, time_last_modified float)
    """
    if specs is None:
        specs = []
    global _total_counter, _ignored_dirs_counter, _ignored_files_counter
    entries = []
    dirs, files = None, None
    for _, dirs, files in os.walk(root, followlinks=follow_symlinks):
        # probably the followlinks doesn't do anything here since we do not recurse with os.walk
        break

    def sorter(input_list):
        if input_list is None:
            return None
        if reverse:
            return reversed(sorted(input_list))
        return sorted(input_list)

    dirs = sorter(dirs)
    files = sorter(files)
    if dirs is not None and recursive:
        abs_dirs = []
        for d in dirs:
            abs_dir = (root / d).absolute()
            if not abs_dir.is_dir():  # protect against things like mounts, only index real dirs.
                continue
            if abs_dir.is_symlink():
                if not follow_symlinks:
                    continue
                else:
                    # reading the link will raise an error if the link is broken.
                    abs_dir.readlink()
            abs_dirs.append(abs_dir)
        if not follow_symlinks:
            abs_dirs = [d for d in abs_dirs if not d.is_symlink()]
        if len(specs) > 0:
            # convert to relative directories and filter with the pathspecs, then convert back
            rel_dirs = [d.relative_to(base_root) for d in abs_dirs]
            n_before = len(rel_dirs)
            rel_dirs = list(apply_pathspecs(rel_dirs, specs))
            _ignored_dirs_counter += n_before - len(rel_dirs)
            abs_dirs = [base_root / d for d in rel_dirs]
        for root_new in abs_dirs:
            entries += _recursive_index(
                root_new,
                base_root,
                depth + 1,
                recursive=recursive,
                verbose=verbose,
                reverse=reverse,
                show_file_if_verbose=show_file_if_verbose,
                specs=specs,
                follow_symlinks=follow_symlinks,
            )
    if files is None:
        return entries

    abs_files = []
    for f in files:
        abs_file = (root / f).absolute()
        if abs_file.is_symlink() and follow_symlinks:
            # pathlib silently ignores recursive symlinks. the way to detect the error is because
            # it is not a directory anymore, but following it will lead to a directory.
            linked_file = abs_file.readlink()
            if linked_file.is_dir() and not abs_file.is_dir():
                raise RuntimeError(f"Broken symlink, potentially recursive: {abs_file} -> {linked_file}")
        if not abs_file.is_file():
            # .is_file() safeguards against stuff like /dev/zero which returns .is_char_device()
            continue
        if abs_file.is_symlink() and not follow_symlinks:
            continue
        abs_files.append(abs_file)
    rel_files = [f.relative_to(base_root) for f in abs_files]
    if len(specs) > 0:
        # filter with pathspecs on relative file level
        n_before = len(rel_files)
        rel_files = list(apply_pathspecs(rel_files, specs))
        _ignored_files_counter += n_before - len(rel_files)
        abs_files = [base_root / ff for ff in rel_files]

    for abs_file, rel_file in zip(abs_files, rel_files):
        rel_file_str = rel_file.as_posix()

        # get size and mod time
        stat = abs_file.stat()
        entries.append((rel_file_str, int(stat.st_size), float(stat.st_mtime)))

        # logging
        if verbose and _total_counter % _count_every == 0:
            if show_file_if_verbose:
                file_out = rel_file_str
                if len(rel_file_str) > _max_print:
                    half = _max_print // 2
                    file_out = " ".join((rel_file_str[: half - 3], "...", rel_file_str[half:]))
            else:
                file_out = f"{stat.st_size / 1024 ** 2:13,.3f} MB"
            _pbar.set_description(f" {next(_spinner)} indexing {file_out}", refresh=False)
        _pbar.update(1)
        _total_counter += 1
    return entries



import os
import sys
from pathlib import Path
from typing import Optional

from packg.typext import PathType


def navigate_to_git_root(starting_dir: Optional[PathType] = None, verbose=False):
    target_dir = find_git_root(starting_dir, verbose)
    os.chdir(target_dir)


def find_git_root(starting_dir: Optional[PathType] = None, verbose=False):
    if starting_dir is None:
        starting_dir = os.getcwd()
    if verbose:
        print(f"Starting dir: {starting_dir}")
    current_dir = Path(starting_dir)
    found = False
    for _ in range(128):
        if (current_dir / ".git").is_dir():
            found = True
            break
        if current_dir == current_dir.parent:
            break
        current_dir = current_dir.parent
    if not found:
        raise RuntimeError("Could not find repo root.")
    if verbose:
        print(f"Final dir:    {current_dir}")
    return current_dir


def add_git_root_to_path(starting_dir: Optional[PathType] = None, verbose=False):
    target_dir = find_git_root(starting_dir, verbose).as_posix()
    if target_dir != starting_dir and target_dir not in sys.path:
        sys.path.insert(0, target_dir)
        if verbose:
            print(f"Added to path: {target_dir}")



from importlib import reload, import_module
from types import ModuleType


def reload_recursive(module, reload_all=False, verbose=False):
    _reload(module, reload_all, set(), verbose=verbose)


def _reload(module, reload_all, reloaded, verbose=False):
    if verbose:
        print(f"trying to reload {module.__name__}")
    if isinstance(module, ModuleType):
        module_name = module.__name__
    elif isinstance(module, str):
        module_name, module = module, import_module(module)
    else:
        raise TypeError(
            "'module' must be either a module or str; " f"got: {module.__class__.__name__}"
        )

    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        # is it a module, has it already been reloaded, is it a proper submodule or reload_all
        if (
            isinstance(attr, ModuleType)
            and attr.__name__ not in reloaded
            and (reload_all or attr.__name__.startswith(module_name))
        ):
            if verbose:
                print(f"recurse into {attr.__name__}")
            _reload(attr, reload_all, reloaded)

    if verbose:
        print(f"reloading module: {module.__name__}")
    reload(module)
    reloaded.add(module_name)



import logging
from timeit import default_timer as timer

import tomlkit
from tomlkit import TOMLDocument

from packg.iotools.file_reader import read_text_from_file_or_io
from packg.typext import PathOrIO


def load_toml(file_or_io: PathOrIO, verbosity: int = logging.WARNING) -> TOMLDocument:
    start_timer = timer()
    data_str = read_text_from_file_or_io(file_or_io)
    try:
        data_toml = loads_toml(data_str)
    except Exception as e:
        print(f"Error loading toml file {file_or_io}: {e}")
        raise e
    if verbosity <= logging.INFO:
        print(f"Loaded json file {file_or_io} in {timer() - start_timer:.3f} seconds")
    return data_toml


def loads_toml(s: str) -> TOMLDocument:
    source_toml_dict: TOMLDocument = tomlkit.loads(s)
    return source_toml_dict


# source_toml_str_re = tomlkit.dumps(source_toml_dict)


# # todo
# def dump_toml():
#     raise NotImplementedError()
#
#
# def dumps_toml():
#     raise NotImplementedError()



"""
Some reusable fixtures and example setups for writing the tests.
"""

import threading
from functools import partial
from http import server
from pathlib import Path

import pytest

BIN_FILE = "test.bin"
TXT_FILE = "test.test"
BIN_CONTENT = b"test" * 1000
TXT_CONTENT = "ütest" * 1000


@pytest.fixture(scope="session")
def webserver(tmp_path_factory: pytest.TempPathFactory):
    tmp_path = tmp_path_factory.mktemp("test_io_web")
    (Path(tmp_path) / BIN_FILE).write_bytes(BIN_CONTENT)
    (Path(tmp_path) / TXT_FILE).write_text(TXT_CONTENT, encoding="utf-8")
    print(f"Created path for temporary webserver: {tmp_path}")

    address = ("localhost", 0)  # port=0 means to select an arbitrary unused port
    handler_class = partial(server.SimpleHTTPRequestHandler, directory=str(tmp_path))
    http_server = server.ThreadingHTTPServer(address, handler_class)

    host, port = http_server.socket.getsockname()[:2]
    print(f"Serving HTTP on (http://{host}:{port}/) ...")

    thread = threading.Thread(target=http_server.serve_forever)
    thread.daemon = True
    thread.start()

    yield host, port

    print(f"Closing server...")
    http_server.shutdown()
    thread.join()



import urllib.parse


def quote_with_urlparse(sentence: str, prefix: str = "") -> str:
    """
    Quote the input s.t. it is safe for using in a URL.
    Useful also for e.g. using the input
    as file names.

    Args:
        sentence: input string
        prefix: prefix to add to the quoted string, useful to avoid empty return strings

    Returns:
        quoted string
    """

    quoted = urllib.parse.quote(sentence, safe="")
    return f"{prefix}{quoted}"


def unquote_with_urlparse(sentence: str, prefix: str = "") -> str:
    sentence_no_prefix = sentence[len(prefix) :]
    return urllib.parse.unquote(sentence_no_prefix)



"""
Wrapper functions for YAML I/O.
"""
import difflib
import os
import sys
from collections import abc
from pathlib import Path
from typing import Any

import yaml
from typedparser.objects import is_any_mapping, is_any_iterable, compare_nested_objects,\
    modify_nested_object

from packg.iotools.file_reader import read_text_from_file_or_io
from packg.typext import PathOrIO, PathTypeCls


def load_yaml(file_or_io: PathOrIO) -> Any:
    yaml_str = read_text_from_file_or_io(file_or_io)
    return loads_yaml(yaml_str)


def loads_yaml(yaml_str: str) -> Any:
    return yaml.load(yaml_str, Loader=yaml.SafeLoader)


def dump_yaml(
    obj: Any,
    file_or_io: PathOrIO,
    standard_format: bool = True,
    check_roundtrip: bool = True,
    create_parent=False,
    **kwargs,
) -> None:
    """
    convert python object to yaml string and write to file. see dumps_yaml for details.
    """
    s = dumps_yaml(obj, standard_format=standard_format, check_roundtrip=check_roundtrip, **kwargs)
    if isinstance(file_or_io, PathTypeCls):
        if create_parent:
            os.makedirs(Path(file_or_io).parent, exist_ok=True)
        Path(file_or_io).write_text(s, encoding="utf8")
        return
    file_or_io.write(s)


def _path_to_str_if_path(x):
    if isinstance(x, Path):
        return x.as_posix()
    return x


def dumps_yaml(
    obj: Any, standard_format: bool = True, check_roundtrip: bool = True, **kwargs
) -> str:
    """
    convert python object to yaml string

    Args:
        obj:
        standard_format:
            if True, use the standard yaml format
                (multi-line lists and dicts, default 2 space indents).
            if False, uses custom format (compatible with the standard format):
                - lists and dicts inside those lists are put on a single line
                - default 4 space indents indents of 2
                - strings are put in quotes
                - keys are not sorted by default, but the order is kept
        check_roundtrip: if True, check that the object can be reconstructed from the yaml string.
            this will cost performance if the object is very large.
        **kwargs: passed to yaml dumper if standard_format is True

    Returns:
        yaml string
    """
    if is_any_mapping(obj):
        # yaml does not understand pathlib.Path so first of all convert all paths to str
        obj = modify_nested_object(obj, _path_to_str_if_path, return_copy=True)
    if standard_format:
        return yaml.dump(obj, Dumper=yaml.SafeDumper, **kwargs)
    yaml_str = _dumps_yaml_recursive(obj, **kwargs)
    if check_roundtrip:
        re_obj = loads_yaml(yaml_str)

        if re_obj != obj:
            print(f"---------- Original object:\n{obj}\n", file=sys.stderr)
            print(f"---------- Reconstructed object:\n{re_obj}\n", file=sys.stderr)
            diffs = compare_nested_objects(obj, re_obj)
            diff_str = "\n".join(diffs)
            print(f"---------- Diffs:\n{diff_str}\n", file=sys.stderr)
            raise RuntimeError(
                "roundtrip failed (original object cannot be reconstructed from yaml, see stderr)"
            )
    return yaml_str


def _dumps_yaml_recursive(
    obj: Any, indent: int = 4, _indent_level: int = 0, _is_inside_list: bool = False
) -> str:
    """
    Modified version of yaml.dump with abbreviated lists and dicts.
    """
    # setup key-value collector and indent level
    indent = " " * (_indent_level * indent)

    # check type
    if isinstance(obj, (bool, int, float, type(None))):
        # by yaml standard, if there is only a single objects in the file, append "..." (eod)
        return _convert_single_object_to_yaml_str(
            obj, remove_eod=_indent_level > 0 or _is_inside_list
        )
    if isinstance(obj, str):
        # put quotes around strings
        return f'"{obj}"'
    if is_any_mapping(obj):
        if _is_inside_list:
            # short: dicts inside lists will be kept in the abbreviated form: {key: "value"}
            if len(obj) == 0:
                return "{}"
            dct_strs = ["{"]
            for sub_key, sub_val in obj.items():
                dct_strs.append(
                    f"{sub_key}: " f"{_dumps_yaml_recursive(sub_val, _is_inside_list=True)}"
                )
                dct_strs.append(", ")
            dct_strs.pop()
            dct_strs.append("}")
            return "".join(dct_strs)
        # long: build the dict line by line
        ret_list = []
        for k, v in obj.items():
            kv_sep = "\n" if isinstance(v, abc.Mapping) else " "
            recursive_result = _dumps_yaml_recursive(
                v, _indent_level=_indent_level + 1, _is_inside_list=_is_inside_list
            )
            ret_list.append(f"{indent}{k}:{kv_sep}{recursive_result}")
        return "\n".join(ret_list)
    if is_any_iterable(obj):
        # iterate lists
        return f"[{', '.join((_dumps_yaml_recursive(v, _is_inside_list=True) for v in obj))}]"
    raise ValueError(f"dict to yaml, value type not understood: type={type(obj)} value={obj}")


def _convert_single_object_to_yaml_str(obj, remove_eod: bool = True):
    yaml_str = yaml.dump(obj, Dumper=yaml.SafeDumper, default_flow_style=True)
    if remove_eod:
        # yaml appends "..." (end of document) to the end of the string, remove
        yaml_str = yaml_str.split("...")[0].strip()
    return yaml_str



"""

"""

from collections import defaultdict
from pathlib import Path

from attrs import define
from loguru import logger

from packg.iotools import make_index
from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    folder: Path = add_argument("folder", type=str, help="Directory to check")
    ignore_case: bool = add_argument(shortcut="-i", action="store_true", help="Ignore case")


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    folder = Path(args.folder)
    if not folder.is_dir():
        raise ValueError(f"Not a dir: {folder}")

    file_index = make_index(folder)
    endings = defaultdict(int)
    for filename in file_index.keys():
        splits = Path(filename).name.split(".")
        if len(splits) == 1:
            ending = ""
        else:
            ending = splits[-1]
        if args.ignore_case:
            ending = ending.lower()
        endings[ending] += 1
    print(f"{endings}")


if __name__ == "__main__":
    main()



from __future__ import annotations

import time
import types
from collections import defaultdict
from multiprocessing import Process, Queue
from statistics import mean
from timeit import default_timer
from typing import Optional, Any
from attr import define, field
from loguru import logger

from packg.dtime import format_seconds_adaptive
from packg.stats import AvgMetric
from packg.tqdmext import tqdm_max_ncols


@define
class MultiProcessorProducerConsumer:
    """
    Run P producers on some input, pass their output to N consumers, and aggregate their output.

    Usage:
        1) specify producer and consumer. Either as a simple function or inheriten Producer/Consumer

        - define producer function that takes input_args and produces output_args
        - subclass Consumer and define setup, __call__, complete. __call__ takes output_args
        - create instance. (optionally set n_producers = n_consumers = 0 for foreground execution)
        - loop put(input_args) some tasks
        - outputs = run()  # list of length n_consumers with the return from complete()

    todo
        Improve error handling with error queues if needed
        Improve passing dict of flags / config, instead of one bool for each config item
        Create another example in this file
    """

    n_producers: int
    n_consumers: int
    producer_class_or_fn: callable
    consumer_class_or_fn: callable

    producer_args: tuple = field(factory=tuple)
    producer_kwargs: dict = field(factory=dict)
    consumer_args: tuple = field(factory=tuple)
    consumer_kwargs: dict = field(factory=dict)
    ignore_errors: bool = False
    verbose: bool = False
    total: Optional[int] = None
    pbar_desc: str = "Multiprocessing"

    producer_class: Any = field(init=False)
    consumer_class: Any = field(init=False)
    producer_worker_list: list[Process] = field(factory=list, init=False)
    producer_object_list: list[Any] = field(factory=list, init=False)
    consumer_worker_list: list[Process] = field(factory=list, init=False)
    consumer_object_list: list[Any] = field(factory=list, init=False)
    q_in: Optional[Queue] = field(init=False, default=None)
    q_producer_out: Optional[Queue] = field(init=False, default=None)
    q_consumer_out: Optional[Queue] = field(init=False, default=None)
    q_timer: Optional[Queue] = field(init=False, default=None)
    pbar: tqdm_max_ncols = field(init=False)
    start_time: int = field(init=False)
    processed: int = field(init=False)
    input_counter: int = field(init=False)
    is_foreground: bool = field(init=False, default=True)
    wait_times_put: AvgMetric = field(init=False, default=AvgMetric())

    def __attrs_post_init__(self):
        assert self.n_producers >= 0, f"n_producers={self.n_producers} but must be >= 0"
        assert self.n_consumers >= 0, f"n_consumers={self.n_consumers} but must be >= 0"

        self._setup_worker_classes()
        is_foreground_producer = self.n_producers == 0
        is_foreground_consumer = self.n_consumers == 0
        if is_foreground_consumer != is_foreground_producer:
            raise ValueError(
                f"n_consumers={self.n_producers} and n_consumers={self.n_consumers} but must be "
                f"either both 0 (running in foreground) or both > 0 (running in background)."
            )

        if is_foreground_producer and is_foreground_consumer:
            self.producer_object_list = [self.producer_class(0)]
            self.producer_object_list[0].setup(*self.producer_args, **self.producer_kwargs)
            self.consumer_object_list = [self.consumer_class(0)]
            self.consumer_object_list[0].setup(*self.consumer_args, **self.consumer_kwargs)
        else:
            self.q_in = Queue(maxsize=self.n_producers * 2)
            self.q_producer_out = Queue(maxsize=self.n_producers * 2)
            self.q_consumer_out = Queue(maxsize=0)
            self.q_timer = Queue(maxsize=0)
            self.is_foreground = False

        for n in range(self.n_producers):
            producer_object = self.producer_class(n)
            producer_object.setup(*self.producer_args, **self.producer_kwargs)
            self.producer_object_list.append(producer_object)
            w = Process(
                target=producer_object.multi_fn_producer,
                args=(self.q_in, self.q_producer_out, self.q_timer, self.ignore_errors),
            )
            w.start()
            self.producer_worker_list.append(w)

        for n in range(self.n_consumers):
            consumer_object = self.consumer_class(n)
            consumer_object.setup(*self.consumer_args, **self.consumer_kwargs)
            self.consumer_object_list.append(consumer_object)
            w = Process(
                target=consumer_object.multi_fn_consumer,
                args=(self.q_producer_out, self.q_consumer_out, self.q_timer, self.ignore_errors),
            )
            w.start()
            self.consumer_worker_list.append(w)

        self.pbar = tqdm_max_ncols(total=self.total, desc=self.pbar_desc)
        self.start_time = default_timer()
        self.processed = 0
        self.input_counter = 0

    def _setup_worker_classes(self):
        if isinstance(self.consumer_class_or_fn, types.FunctionType):
            # detected consumer function
            assert len(self.consumer_args) == 0, (
                f"{len(self.consumer_args)} consumer_args given but must be empty "
                f"when using a consumer function."
            )
            assert len(self.consumer_kwargs) == 0, (
                f"{len(self.consumer_kwargs)} consumer_kwargs given but must be empty "
                f"when using a consumer function."
            )
            self.consumer_class = SimpleConsumer
            self.consumer_args = (self.consumer_class_or_fn,)
        else:
            # detected consumer class
            self.consumer_class = self.consumer_class_or_fn

        if isinstance(self.producer_class_or_fn, types.FunctionType):
            # detected producer function
            assert len(self.producer_args) == 0, (
                f"{len(self.producer_args)} producer_args given but must be empty "
                f"when using a producer function."
            )
            assert len(self.producer_kwargs) == 0, (
                f"{len(self.producer_kwargs)} producer_kwargs given but must be empty "
                f"when using a producer function."
            )
            self.producer_class = SimpleProducer
            self.producer_args = (self.producer_class_or_fn,)
        else:
            # detected producer class
            self.producer_class = self.producer_class_or_fn

    def _debug_print(self, msg):
        if self.verbose:
            self.pbar.write(msg)

    def put(self, *args):
        if self.is_foreground:
            output = self.producer_object_list[0](*args)
            output = _ensure_fn_output_is_tuple(output)
            self.consumer_object_list[0](*output)
            self.update_pbar(1)
            return

        start_time = default_timer()
        self.q_in.put(args)
        self.wait_times_put.update(default_timer() - start_time)

        self.input_counter += 1
        # with W workers, the first W will be immediately accepted. 2*W will be queued.
        # so only start counting after 3*W for an accurate pbar.
        if self.input_counter > self.n_producers * 3:
            self.update_pbar(1)

    def update_pbar(self, amount=1):
        # if there are more producers than queue items, safeguard pbar against overflow
        if self.total is not None:
            amount = min(amount, self.total - self.processed)
        if amount <= 0:
            return

        self.processed += amount
        sec_per_iter = (default_timer() - self.start_time) / max(self.processed, 1)
        desc_post = f" {sec_per_iter:10.2f} s/it"
        if sec_per_iter < 1:
            iter_per_sec = 1 / sec_per_iter
            desc_post = f" {iter_per_sec:10.2f} it/s"
        if self.total is None:
            rem_time_fmt = f""
        else:
            rem_time = sec_per_iter * (self.total - self.processed)
            rem_time_fmt = format_seconds_adaptive(rem_time, "{:5.2f}{}")
        self.pbar.set_description(f"{self.pbar_desc}{desc_post} {rem_time_fmt} left", refresh=False)
        self.pbar.update(amount)

    def run(self):
        if self.is_foreground:
            # already done
            self.pbar.close()
            return [self.consumer_object_list[0].complete()]

        self._debug_print(f"Add {self.n_producers} term signals for producers to input queue")
        for _ in range(self.n_producers):
            start_time = default_timer()
            self.q_in.put(None)
            self.wait_times_put.update(default_timer() - start_time)

        remaining = self.q_in.qsize()
        self._debug_print(f"Wait for producers to read input queue ({remaining} rem.)")
        wait_counter = 0
        delay = 0.2
        while not self.q_in.empty():
            if wait_counter % 5 == 0:
                now_remaining = self.q_in.qsize()
                if now_remaining < remaining:
                    delta = remaining - now_remaining
                    remaining = now_remaining
                    self.update_pbar(delta)
            wait_counter += 1
            time.sleep(delay)
        self.update_pbar(remaining)

        remaining = self.q_producer_out.qsize()
        self._debug_print(f"Wait for consumers to read producer output queue ({remaining} rem.)")
        while not self.q_producer_out.empty():
            time.sleep(delay)

        self._debug_print(f"Add {self.n_consumers} term signals for consumers to input queue")
        for _ in range(self.n_consumers):
            self.q_producer_out.put(None)

        self._debug_print(f"Collect data from consumer output queue")
        outputs = []
        for _ in range(self.n_consumers):
            out = self.q_consumer_out.get()
            outputs.append(out)

        # note: all queues have to be empty before workers can be joined otherwise this will hang
        self._debug_print(f"Joining workers")
        for w in self.producer_worker_list:
            w.join()
            self.update_pbar(1)
            w.terminate()
        for w in self.consumer_worker_list:
            w.join()
            w.terminate()

        m_wait_put = self.wait_times_put.avg

        times = defaultdict(list)
        for _ in range(self.n_consumers + self.n_producers * 2):
            time_k, _time_i, time_v = self.q_timer.get()
            times[time_k].append(time_v)

        time_str = ", ".join([f"{k}: {mean(vs):.1f}s" for k, vs in times.items()])

        self.pbar.close()
        self.q_in.close()
        self.q_producer_out.close()
        self.q_consumer_out.close()
        self.q_timer.close()
        total_speed = (default_timer() - self.start_time) / max(self.processed, 1)
        self.pbar.write(
            f"Average wait: main put: {m_wait_put:.2f}s, {time_str}. "
            f"Final speed: {total_speed:.3f}s/it"
        )

        return outputs


def _ensure_fn_output_is_tuple(output):
    if not isinstance(output, tuple):
        output = (output,)
    return output


class Producer:
    def __init__(self, i: int):
        self.i = i
        self.wait_times_get = AvgMetric()
        self.wait_times_put = AvgMetric()

    def setup(self, *args, **kwargs) -> None:
        pass

    def __call__(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def multi_fn_producer(
        self,
        q_in: Queue,
        q_producer_out: Queue,
        q_timer: Queue,
        ignore_errors: bool,
    ):
        while True:
            start_time = default_timer()
            args = q_in.get()
            self.wait_times_get.update(default_timer() - start_time)
            if args is None:
                break
            try:
                # producer_fn: callable,
                out = self(*args)
            except Exception as e:
                if ignore_errors:
                    logger.error(f"Error in multi_function: {e}")
                    q_producer_out.put(e)
                    continue
                raise e
            out = _ensure_fn_output_is_tuple(out)
            start_time = default_timer()
            q_producer_out.put(out)
            delta = default_timer() - start_time
            self.wait_times_put.update(delta)

        q_timer.put((f"producer-get", self.i, self.wait_times_get.avg))
        q_timer.put((f"producer-put", self.i, self.wait_times_put.avg))


class SimpleProducer(Producer):
    def setup(self, producer_fn: callable):
        self.producer_fn = producer_fn

    def __call__(self, *args, **kwargs) -> None:
        return self.producer_fn(*args, **kwargs)


class Consumer:
    def __init__(self, i: int):
        self.i = i
        self.wait_times_get = AvgMetric()

    def setup(self, *args, **kwargs) -> None:
        pass

    def __call__(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def complete(self) -> Any:
        raise NotImplementedError

    def multi_fn_consumer(
        self, q_producer_out: Queue, q_consumer_out: Queue, q_timer: Queue, ignore_errors: bool
    ):
        while True:
            start_time = default_timer()
            args = q_producer_out.get()
            self.wait_times_get.update(default_timer() - start_time)
            if args is None:
                break
            try:
                self(*args)
            except Exception as e:
                if ignore_errors:
                    logger.error(f"Error in multi_function: {e}")
                else:
                    raise e
        final_output = self.complete()
        # print(f"Putting output to consumer out q: {type(final_output)=} {len(final_output)=}")
        # q_consumer_out never blocks so no reason to time it.
        q_consumer_out.put(final_output)
        q_timer.put((f"consumer-get", self.i, self.wait_times_get.avg))


class SimpleConsumer(Consumer):
    def setup(self, consumer_fn: callable):
        self.collected_outputs = []
        self.consumer_fn = consumer_fn

    def __call__(self, *args, **kwargs):
        self.collected_outputs.append(self.consumer_fn(*args, **kwargs))

    def complete(self):
        return self.collected_outputs


def fn_produce_one_value(t_x: int):
    time.sleep(t_x / 10)
    return t_x * 2


def fn_consume_one_value(t_y: int):
    time.sleep(t_y / 20)
    return t_y


def main():
    ix = list(range(10))
    proc = MultiProcessorProducerConsumer(
        30, 2, fn_produce_one_value, fn_consume_one_value, total=10, pbar_desc="Test", verbose=True
    )
    for in_x in ix:
        proc.put(in_x)
    outputs = proc.run()
    print(outputs)
    print("Done")


if __name__ == "__main__":
    main()



from __future__ import annotations

import math
import numpy as np
from typing import Iterable, Union, Optional


def clip_rectangle_coords(rectangle_coords: tuple[int, int, int, int], w: int, h: int):
    """

    Args:
        rectangle_coords: (x1, y1, x2, y2) of a rectangle
        w: image width
        h: image height

    Returns:
        tuple of rectangle coordinates clipped to image size
    """
    x1, y1, x2, y2 = rectangle_coords
    x1 = max(0, x1)
    y1 = max(0, y1)
    x2 = min(w - 1, x2)
    y2 = min(h - 1, y2)
    return x1, y1, x2, y2


def round_half_up(x: float) -> int:
    """
    Python/numpy do bankers rounding (round to even). This method rounds 0.5 up to 1 instead.
    """
    return int(math.ceil(x - 0.5))


def round_half_down(x: float) -> int:
    return int(math.floor(x + 0.5))


def np_round_half_up(array: Union[np.ndarray, Iterable[Union[int, float]]]):
    if not isinstance(array, np.ndarray):
        array = np.array(array)
    return np.floor(array + 0.5)


def np_round_half_down(array: Union[np.ndarray, Iterable[Union[int, float]]]):
    if not isinstance(array, np.ndarray):
        array = np.array(array)
    return np.ceil(array - 0.5)


def convert_unsigned_int_to_bytes(int_input: int, length: int = 4) -> bytes:
    """
    convert integer to bytes

    Args:
        int_input:
        length: 4 = int32 (max 4B), 8 = int64 (max 1.9e19)

    Returns:
        bytes
    """
    return int(int_input).to_bytes(length, "big")


def convert_bytes_to_unsigned_int(bytes_input: bytes) -> int:
    """
    convert bytes to integer

    Args:
        bytes_input:

    Returns:
        integer
    """
    if len(bytes_input) == 0:
        raise ValueError("bytes_input must have length > 0 but is empty")
    return int.from_bytes(bytes_input, "big")


def np_str_len(str_arr: Union[np.ndarray, Iterable[str]]) -> np.ndarray:
    """
    Fast way to get string length in a numpy array with datatype string.

    Args:
        str_arr: Numpy array of strings with arbitrary shape.

    Returns:
        Numpy array of string lengths, same shape as input.

    Notes:
        Source: https://stackoverflow.com/questions/44587746/length-of-each-string-in-a-numpy-array
        The latest improved answers don't really work. This code should work for all except strange special characters.
    """
    if not isinstance(str_arr, np.ndarray):
        # also support iterables of strings
        str_arr = np.array(str_arr)
    # check input type
    if str(str_arr.dtype)[:2] != "<U":
        raise TypeError(
            f"Computing string length of dtype {str_arr.dtype} will not work correctly. Cast array to string first."
        )

    # see the link in the docstring as an explanation of what exactly is happening here
    try:
        v = str_arr.view(np.uint32).reshape(str_arr.size, -1)
    except TypeError as e:
        print(f"Input {str_arr} shape {str_arr.shape} dtype {str_arr.dtype}")
        raise e
    len_arr = np.argmin(v, 1)
    len_arr[v[np.arange(len(v)), len_arr] > 0] = v.shape[-1]
    len_arr = np.reshape(len_arr, str_arr.shape)
    return len_arr


def format_float_maybe_scientific(
    in_float,
    decimals: Optional[int] = None,
    integers: Optional[int] = None,
    lower_limit: float = 1e-4,
    upper_limit: float = 1e6,
) -> str:
    """
    Format float as string, as decimal notation if with limits, scientific otherwise.

    Args:
        in_float:
        decimals: how many numbers after decimal point
        integers: how many numbers before decimal point
        lower_limit: switch to scientific notation below this
        upper_limit: switch to scientific notation above this

    Returns:

    """
    in_float = float(in_float)
    if decimals is not None:
        decimals_str = int(decimals)
        assert decimals_str >= 0
    else:
        decimals_str = ""

    if integers is not None:
        total_len_str = int(integers)
        assert total_len_str >= 0
    else:
        total_len_str = ""

    len_and_dec_str = f"{total_len_str}.{decimals_str}" if decimals_str != "" else ""

    format_type = "e" if abs(in_float) < lower_limit or abs(in_float) > upper_limit else "f"
    format_str = f"{{:{len_and_dec_str}{format_type}}}"
    return format_str.format(in_float)



"""
Utilities to read content of a single file.
"""
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Union, Iterable

from packg.typext import PathOrIO, PathTypeCls, PathType


@contextmanager
def open_file_or_io(
    file_or_io: PathOrIO,
    mode="r",
    encoding="utf-8",
    create_parent=False,
):
    should_close = False
    if isinstance(file_or_io, PathTypeCls):
        file_or_io = Path(file_or_io)
        if create_parent:
            os.makedirs(file_or_io.parent, exist_ok=True)
        if "b" in mode:
            encoding = None
        fh = file_or_io.open(mode, encoding=encoding)
        should_close = True
    else:
        fh = file_or_io
    yield fh
    if should_close:
        fh.close()


def read_text_from_file_or_io(file_or_io: PathOrIO, encoding: str = "utf-8") -> str:
    """
    Args:
        file_or_io: file name or open file-like object
        encoding: encoding to use for reading

    Returns:
        text content
    """
    if isinstance(file_or_io, PathTypeCls):
        return Path(file_or_io).read_text(encoding=encoding)
    return file_or_io.read()


def read_bytes_from_file_or_io(file_or_io: PathOrIO) -> bytes:
    """

    Args:
        file_or_io: file name or open file-like object

    Returns:
        bytes content
    """
    if isinstance(file_or_io, PathTypeCls):
        return Path(file_or_io).read_bytes()
    return file_or_io.read()


def yield_chunked_bytes(file_or_io: PathOrIO, chunk_size=1024 * 1024) -> Iterable[bytes]:
    """

    Args:
        file_or_io: file name or open file-like object
        chunk_size: chunk size in bytes, default 1MB

    Returns:
        bytes content
    """
    with open_file_or_io(file_or_io, mode="rb") as fh:
        while True:
            data = fh.read(chunk_size)
            if len(data) == 0:
                break
            yield data


def yield_lines_from_object(
    lines_obj: Union[str, Iterable[str]], strip: bool = True, skip_empty: bool = True
) -> Iterable[str]:
    """
    Read lines from input, strip whitespaces, skip empty lines, yield lines.

    Args:
        lines_obj: Either str or iterable of str (list, opened file)
        strip: strip whitespace from lines
        skip_empty: skip empty lines

    Returns:
        Generator of stripped lines

    Examples:
        >>> for li in yield_lines_from_object(["  a  ", "  ", "  b  "]): print(li, end=",")
        a,b,
    """
    if isinstance(lines_obj, str):
        lines_obj = lines_obj.splitlines()
    for line in lines_obj:
        if strip:
            line = line.strip()
        if skip_empty:
            if line == "":
                continue
        yield line


def yield_lines_from_file(
    file: PathType, strip: bool = True, skip_empty: bool = True, encoding: str = "utf-8"
) -> Iterable[str]:
    """
    Read lines from input, strip whitespaces, skip empty lines, yield lines.

    Args:
        file: file path as string
        strip: strip whitespace from lines
        skip_empty: skip empty lines
        encoding: encoding to use for reading

    Returns:
        Generator of stripped lines
    """
    content = Path(file).read_text(encoding=encoding)
    yield from yield_lines_from_object(content, strip=strip, skip_empty=skip_empty)



"""
Remove build artifacts and other clutter from a directory.
"""
import shutil
from pathlib import Path
from typing import Optional

from loguru import logger

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from typedparser import VerboseQuietArgs, add_argument, TypedParser
from attrs import define


@define
class Args(VerboseQuietArgs):
    folder: Path = add_argument("folder", type=str, help="Directory to check")
    write: bool = add_argument(shortcut="-w", action="store_true", help="Confirm removing files")
    clean_packages: bool = add_argument(
        shortcut="-p", action="store_true", help="remove package builds and dist"
    )
    clean_pycache: bool = add_argument(
        shortcut="-c", action="store_true", help="remove all pycache and pytests cache"
    )
    clean_all: bool = add_argument(shortcut="-a", action="store_true", help="remove everything")
    remove_empty_dirs: bool = add_argument(
        shortcut="-e", action="store_true", help="remove empty directories"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    base_dir = args.folder.resolve()
    logger.warning(f"Cleaning up: {base_dir}")

    def rmtree_verbose(path):
        logger.info(f"RM {path}")
        if args.write:
            shutil.rmtree(path, ignore_errors=True)

    glob_strs = []
    if args.clean_packages or args.clean_all:
        glob_strs = ["**/build", "**/dist", "**/*.egg-info"]

    if args.clean_pycache or args.clean_all:
        glob_strs += ["**/__pycache__", "**/.pytest_cache", "**/*.pyc", "**/*.pyo"]

    if args.clean_all:
        glob_strs += ["**/.ipynb_checkpoints"]

    logger.info(f"Searching for {glob_strs}")
    n_removed = 0
    for glob_str in glob_strs:
        for file in base_dir.glob(glob_str):
            rmtree_verbose(file)
            n_removed += 1

    logger.info(f"Removed {n_removed} files or dirs")
    if len(glob_strs) == 0 and not args.remove_empty_dirs:
        logger.warning("No operation specified")

    if args.remove_empty_dirs or args.clean_all:
        remove_empty_dirs(base_dir, write=args.write)

    if not args.write:
        logger.warning("Test run (no -w), did not do anything.")


def remove_empty_dirs(folder, level=0, write=False):
    # recursively delete empty subfolders
    subfolders = [f for f in list(folder.glob("*")) if f.is_dir() and not f.is_symlink()]
    for subfolder in subfolders:
        remove_empty_dirs(subfolder, level=level + 1, write=write)

    # if nothing remains, delete this folder
    remaining_elements = list(folder.glob("*"))
    if len(remaining_elements) == 0:
        logger.info(f"RM empty dir: {folder}")
        if write:
            folder.rmdir()


if __name__ == "__main__":
    main()



import os
import pickle

import joblib
from joblib import register_store_backend
from joblib._store_backends import FileSystemStoreBackend  # noqa

from packg.paths import get_cache_dir


class StoreNoNumpy(FileSystemStoreBackend):
    """
    Backend for joblib with pickle instead of numpy_pickle.
    Faster than the original backend for non-numpy objects (e.g. dicts).
    """

    NAME = "no_numpy"

    def load_item(self, path, verbose=1, msg=None):
        full_path = os.path.join(self.location, *path)

        if verbose > 1:
            if verbose < 10:
                print(f"{msg}...")
            else:
                print(f"{msg} from {full_path}")

        mmap_mode = None if not hasattr(self, "mmap_mode") else self.mmap_mode

        filename = os.path.join(full_path, "output.pkl")
        if not self._item_exists(filename):
            raise KeyError(
                f"Non-existing item (may have been cleared).\n" f"File {filename} does not exist"
            )

        assert mmap_mode is None, "Standard pickle does not support mmap_mode"
        with self._open_item(filename, "rb") as fh:
            item = pickle.load(fh)
        return item


register_store_backend(StoreNoNumpy.NAME, StoreNoNumpy)


def get_joblib_memory(
    location=get_cache_dir() / "joblib", verbose=1, numpy_capable=False
) -> joblib.Memory:
    """
    Wrapper for joblib.Memory which uses the StoreNoNumpy backend by default.

    Args:
        location: cache dir
        verbose: higher = more verbose
        numpy_capable: keep False unless needed, it makes loading alot slower for e.g. dicts

    Returns:
        memory: use as @memory.cache decorator for functions
    """
    backend = "local" if numpy_capable else StoreNoNumpy.NAME
    return joblib.Memory(location=location, backend=backend, verbose=verbose)


class SimpleMemoryCache:
    """
    Cache objects in memory.

    Usage:
        # given a function to cache e.g. load_json(file)
        # create cache and call the function with it
        memory_cache = SimpleMemoryCache()
        cache_key = "some_cache_key"
        data = memory_cache.apply_memory_caching(cache_key, load_json, file)
        data_again = memory_cache.apply_memory_caching(cache_key, load_json, file)
        # Function is called once, then data is loaded from memory since cache_key exists.
    """

    def __init__(self):
        self._cache_dict = {}

    def apply_memory_caching(self, cache_key, func, *args, **kwargs):
        cache_dict = self._cache_dict
        if cache_key is None:
            cache_key = joblib.hash((args, kwargs))
        if cache_key not in cache_dict:
            func_result = func(*args, **kwargs)
            cache_dict[cache_key] = func_result
        return cache_dict[cache_key]

    def _load_from_memory_cache(self, cache_key):
        cache_dict = self._cache_dict
        if cache_key not in cache_dict:
            return None
        return cache_dict[cache_key]

    def _write_to_memory_cache(self, cache_key, obj):
        cache_dict = self._cache_dict
        cache_dict[cache_key] = obj


memory_cache = SimpleMemoryCache()



"""
Create a bash autocompletion script for a package or several packages .

Note: If it does not find a module, maybe __init__.py is missing.

see  /usr/share/bash-completion/bash_completion complete -F _longopt complete
to figure out how to complete the 2nd argument same as cat would complete the 1st argument.

"""

from pathlib import Path
from typing import Optional

from attrs import define
from loguru import logger

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.packaging import FILEDIR_AUTOCOMPLETE, create_new_bash_autocomplete_script
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    packages: str = add_argument(shortcut="-p", type=str, default="packg,visiontext")
    target_script: Optional[Path] = add_argument(
        shortcut="-t",
        type=Path,
        help="Target script. If not given, will write to ./autocomplete.sh",
    )
    run_dir: Optional[str] = add_argument(
        shortcut="-r", type=str, help="Only create autocompletion for this directory"
    )
    command_name: str = add_argument(
        shortcut="-c", type=str, default="py", help="Command name to call python"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    packages = args.packages.split(",")
    autocomplete_script = create_new_bash_autocomplete_script(
        packages, run_dir=args.run_dir, command_name=args.command_name
    )
    if args.target_script is None:
        args.target_script = "autocomplete.sh"
    Path(args.target_script).write_text(
        "\n".join([FILEDIR_AUTOCOMPLETE, autocomplete_script]), encoding="utf-8"
    )
    logger.info(f"Created: {args.target_script}")
    logger.debug(autocomplete_script)


if __name__ == "__main__":
    main()



import hashlib
from typing import Any

from packg.iotools import dumps_json
from packg.strings.base64tools import b64_encode_from_bytes


def update_hasher_with_json_bytes(hasher, obj: Any) -> bytes:
    hasher.update(dumps_json(obj).encode("utf-8"))


def hash_object(
    obj: Any, hasher_update_fn=update_hasher_with_json_bytes, hasher_cls=hashlib.sha3_224
) -> str:
    """

    Args:
        obj: any object to create a hash for
        hasher_update_fn: function to update the hash object with the python object
        hasher_cls: class to create the hash object

    Returns:
        hash string. url-safe b64 i.e. a-zA-Z0-9 and _-

    """
    hasher = hasher_cls()
    hasher_update_fn(hasher, obj)
    out_hash_bytes = hasher.digest()
    out_hash = b64_encode_from_bytes(out_hash_bytes, strip_equals=True)
    return out_hash



"""
Todo: create a test case whether this actually triggers on wrong import structure

Utility to import everything from a module and check import path sanity.

In detail: check whether imports inside a library are correctly importing from the source file
and not from the library base __init__.py file. This ensures clean dependencies and avoids
circular imports.

https://github.com/amenck/circular_import_refactor/blob/test-imports-from-source/objects/tests/

Changes:
    - in visit_ImportFrom method replace error with dynamic import, before:
        assert hasattr(module, alias.name)
    - unprivate some names
    - change the import x from y instead of z to a warning since it can be false positive sometimes
        e.g. if the name matches with a public module

Usage:
    import pytest

    from packg.import_from_source import (
        apply_visitor, ImportFromSourceChecker, recurse_modules)

    module_list = list(recurse_modules("packagename", ignore_tests=True, packages_only=False))


    @pytest.mark.parametrize("module", module_list)
    def test_imports_from_source(module: str) -> None:
        print(f"Importing: {module}")
        apply_visitor(module=module, visitor=ImportFromSourceChecker(module))

Notes:
    - Currently does not respect try/except blocks around the imports and throws errors anyway
        - This functionality looks difficult to implement
        - Workaround: pass the list of false positives like this:
          ImportFromSourceChecker(module, module_list_to_ignore_not_found=["module.submodule"])
          Then all ModuleNotFoundErrors where the module name starts with a module in this list
          will be ignored.


"""
import logging
from ast import parse, NodeVisitor, ImportFrom
from importlib import util as import_util, import_module
from importlib.machinery import ModuleSpec
from os import path
from pkgutil import iter_modules
from typing import Any, List, Iterator, Optional


def _is_test_module(module_name: str) -> bool:
    components = module_name.split(".")

    return len(components) >= 2 and components[1] == "tests"


def _is_package(module_spec: ModuleSpec) -> bool:
    return module_spec.origin is not None and module_spec.origin.endswith("__init__.py")


def recurse_modules(
    module_name: str,
    ignore_tests: bool = True,
    packages_only: bool = False,
    ignore_main: bool = False,
) -> Iterator[str]:
    if ignore_tests and _is_test_module(module_name):
        return

    module_spec = import_util.find_spec(module_name)

    if module_spec is not None and module_spec.origin is not None:
        if not (packages_only and not _is_package(module_spec)):
            yield module_name

        for child in iter_modules([path.dirname(module_spec.origin)]):
            if child.ispkg:
                yield from recurse_modules(
                    f"{module_name}.{child.name}",
                    ignore_tests=ignore_tests,
                    packages_only=packages_only,
                )
            if packages_only:
                continue
            if ignore_main and child.name == "__main__":
                continue
            yield f"{module_name}.{child.name}"


class ImportFromSourceChecker(NodeVisitor):
    def __init__(self, module: str, module_list_to_ignore_not_found: Optional[List] = None):
        module_spec = import_util.find_spec(module)
        is_pkg = (
            module_spec is not None
            and module_spec.origin is not None
            and module_spec.origin.endswith("__init__.py")
        )

        self._module = module if is_pkg else ".".join(module.split(".")[:-1])
        self._top_level_module = self._module.split(".")[0]
        self._module_list_to_ignore_not_found = module_list_to_ignore_not_found

    def visit_ImportFrom(self, node: ImportFrom) -> Any:
        # Check that there are no relative imports that attempt to read from a parent module.
        # We've found that there generally is no good reason to have such imports.
        if node.level >= 2:
            raise ValueError(
                f"Import in {self._module} attempts to import from parent module using "
                f"relative import. Please switch to absolute import instead."
            )

        # Figure out which module to import in the case where this is a...
        if node.level == 0:
            # (1) absolute import where a submodule is specified
            assert node.module is not None
            module_to_import: str = node.module
        elif node.module is None:
            # (2) relative import where no module is specified (ie: "from . import foo")
            module_to_import = self._module
        else:
            # (3) relative import where a submodule is specified (ie: "from .bar import foo")
            module_to_import = f"{self._module}.{node.module}"

        # We're only looking at imports of objects defined inside this top-level package
        if not module_to_import.startswith(self._top_level_module):
            return

        # Actually import the module and iterate through all the objects potentially exported by it.
        print(f"    Importing module: {module_to_import}")
        try:
            module = import_module(module_to_import)
        except ModuleNotFoundError as e:
            if self._module_list_to_ignore_not_found is not None:
                for module_to_ignore_not_found in self._module_list_to_ignore_not_found:
                    if module_to_import.startswith(module_to_ignore_not_found):
                        print(f"        Ignore missing module: {module_to_import}")
                        return
            raise e
        for alias in node.names:
            if not hasattr(module, alias.name):
                if alias.name == "*":
                    continue
                attr = import_module(f"{module_to_import}.{alias.name}")
            else:
                attr = getattr(module, alias.name)

            # For some objects (pretty much everything except for classes and functions),
            # we are not able to figure out which module they were defined in...
            # in that case there's not much we can do here, since we cannot
            # easily figure out where we *should* be importing this from in the first place.
            if isinstance(attr, type) or callable(attr):
                try:
                    attribute_module = attr.__module__
                except AttributeError:
                    # e.g. functools.partial outupt does not have __module__ and breaks here
                    continue
            else:
                continue

            # Figure out where we should be importing this class from, and assert that
            # the *actual* import we found matches the place we *should* import from.
            should_import_from = self._get_module_should_import(module_to_import=attribute_module)
            if module_to_import != should_import_from:
                logging.warning(
                    f"(Potential false positive) "
                    f"Imported {alias.name} from {module_to_import}, "
                    f"which is not the public module where this object "
                    f"is defined. Please import from {should_import_from} instead."
                )

    def _get_module_should_import(self, module_to_import: str) -> str:
        """
        This function figures out the correct import path for "module_to_import" from the
        "self._module" module in this instance. The trivial solution here would be to always
        just return "module_to_import", but we want to actually take into account the fact that
        some submodules can be "private" (ie: start with an "_"), in which case we should only
        import from them if self._module is internal to that private module.
        """
        module_components = module_to_import.split(".")
        result: List[str] = []

        for component in module_components:
            if component.startswith("_") and not self._module.startswith(".".join(result)):
                break
            result.append(component)

        return ".".join(result)


def apply_visitor(module: str, visitor: NodeVisitor) -> None:
    module_spec = import_util.find_spec(module)
    assert module_spec is not None
    assert module_spec.origin is not None

    with open(module_spec.origin, "r", encoding="utf-8") as source_file:
        ast = parse(source=source_file.read(), filename=module_spec.origin)

    visitor.visit(ast)



"""
Load and save numpy arrays to bytes and strings.

Note: To get all-in-one
    buffer = io.BytesIO()
    np.save(buffer, score_list)
    arr_bytes = buffer.getvalue()
    buffer = io.BytesIO(cached_score_list)
    score_list = np.load(buffer)
"""

from typing import Tuple

import numpy as np


def dumps_numpy_array(arr) -> Tuple[bytes, str]:
    """
    Convert numpy array into a tuple of bytes and a string describing dtype and shape.
    Useful for then storing it in a database.
    Especially for very small arrays, this method takes much less space then
    np.save or joblib.dump.

    Args:
        arr: Input array

    Returns: tuple of
        bytes: array data as bytes
        str: dtype and shape as string e.g. "float32-100,100" for a 100x100 float32 array

    """
    return arr.tobytes(), f"{arr.dtype.name}-{','.join(str(d) for d in arr.shape)}"


def loads_numpy_array(arr, dtype_shape) -> np.ndarray:
    dtype, shape = dtype_shape.split("-")
    return np.frombuffer(arr, dtype=dtype).reshape([int(s) for s in shape.split(",")])



"""Helper to run system commands and process their output."""

import subprocess
from typing import Tuple, Optional


def systemcall(
    call: str, verbose: bool = False, decode: Optional[str] = "utf-8", shell: bool = True
) -> Tuple[str, str, int]:
    """Run a command with subprocess.Popen and process the output. This call
    is synchronous so output will only returned once the command is done.

    Args:
        call: the command to run
        verbose: verbosity (0=quiet)
        decode: decode the output with this encoding
        shell: run the command in a shell

    Returns:
        stdout, stderr, returncode
    """
    pipe = subprocess.PIPE
    with subprocess.Popen(call, stdout=pipe, stderr=pipe, shell=shell) as process:
        out, err = process.communicate()
        retcode = process.poll()
    if decode is not None:
        out = out.decode(decode)
        err = err.decode(decode)
    if verbose:
        print(f"out {out} err {err} ret {retcode}")
    return out, err, retcode


def assert_command_worked(errmsg: str, cmd: str, out: str, err: str, retcode: int) -> None:
    """Process the output of a systemcall and assert that the command worked.

    Args:
        errmsg: additional error info to display
        cmd: original command that was run (for display purposes)
        out: stdout of the command
        err: stderr of the command
        retcode: return code of the command

    Returns:
        None

    """
    assert retcode == 0, (
        f"command failed: {cmd}\nout: {out}\nerr: {err}\n"
        f"retcode: {retcode}\n"
        f"additional error info: {errmsg}"
    )


def systemcall_with_assert(
    call: str,
    errmsg: str = "none",
    verbose: bool = False,
    decode: Optional[str] = "utf-8",
    shell: bool = True,
) -> Tuple[str, str, int]:
    """Run a command and assert it worked

    Args:
        call: command to run
        errmsg: additional error info to display when the command fails
        verbose: verbosity of the command
        decode: decode the output with this encoding
        shell: run the command in a shell

    Returns:
        stdout, stderr, returncode
    """
    out, err, retcode = systemcall(call, verbose=verbose, decode=decode, shell=shell)
    assert_command_worked(errmsg, call, out, err, retcode)
    return out, err, retcode



import traceback
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull


def format_exception(e, with_traceback=False) -> str:
    error_str, error_name = str(e), type(e).__name__
    if error_str == "":
        out_str = error_name
    else:
        out_str = f"{error_name}: {error_str}"

    if not with_traceback:
        return out_str

    tb_list = traceback.format_tb(e.__traceback__)
    tb_str = "".join(tb_list)
    return f"{tb_str}{out_str}"


@contextmanager
def suppress_stdout_stderr():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w", encoding="utf-8") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield err, out


